import contextlib
import zipfile
from pathlib import Path

import pandas as pd


fields = {
    'site_id': str,
    'patient_id': str,
    'image_id': str,
    'laterality': None,
    'view': None,
    'age': None,
    'cancer': bool,
    'biopsy': bool,
    'invasive': bool,
    'BIRADS': int,
    'implant': bool,
    'density': None,
    'machine_id': str,
    'prediction_id': str,
    'difficult_negative_case': bool,
}


def add_csv_fields(scope):
    def make_loader(field, cast):
        def loader(_row):
            value = _row.get(field)
            if pd.isnull(value):
                return None
            if cast is not None:
                return cast(value)
            return value

        return loader

    for _field, _cast in fields.items():
        scope[_field] = make_loader(_field, _cast)


@contextlib.contextmanager
def unpack(root: str, relative: str):
    unpacked = Path(root) / relative

    if unpacked.exists():
        yield unpacked, True
    else:
        with zipfile.Path(root, relative).open('rb') as unpacked:
            yield unpacked, False



from pathlib import Path

import nibabel as nb
import numpy as np
from connectome import Output, Source, meta
from connectome.interface.nodes import Silent
from deli import load

from .internals import checksum, register
from .utils import deprecate


@register(
    body_region='Head',
    license='EGD data license',
    link='https://xnat.bmia.nl/data/archive/projects/egd',
    modality=('FLAIR', 'MRI T1', 'MRI T1GD', 'MRI T2'),
    prep_data_size='107,49G',
    raw_data_size='40G',
    task='Segmentation',
)
@checksum('egd')
class EGD(Source):
    """
    The Erasmus Glioma Database (EGD): Structural MRI scans, WHO 2016 subtypes,
    and segmentations of 774 patients with glioma [1]_.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.
    version : str, optional
        the data version. Only has effect if the library was installed from a cloned git repository.

    Notes
    -----
    The access to the dataset could be requested at XNAT portal [https://xnat.bmia.nl/data/archive/projects/egd].

    To download the data in the compatible structure we recommend to use
    egd-downloader script [https://zenodo.org/record/4761089#.YtZpLtJBxhF].
    Please, refer to its README for further information.

    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> egd = EGD(root='/path/to/downloaded/data/folder/')
    >>> print(len(egd.ids))
    # 774
    >>> print(egd.t1gd(egd.ids[215]).shape)
    # (197, 233, 189)
    >>> print(egd.manufacturer(egd.ids[444]))
    # Philips Medical Systems

    References
    ----------
    .. [1] van der Voort, Sebastian R., et al. "The Erasmus Glioma Database (EGD): Structural MRI scans,
           WHO 2016 subtypes, and segmentations of 774 patients with glioma."
           Data in brief 37 (2021): 107191.
           https://www.sciencedirect.com/science/article/pii/S2352340921004753

    """

    _root: str = None

    def _base(_root: Silent):
        if _root is None:
            raise ValueError('Please provide the `root` argument')
        return Path(_root)

    @meta
    def ids(_base):
        result = []
        for folder in (_base / 'SUBJECTS').iterdir():
            for suffix in 'FLAIR', 'T1', 'T1GD', 'T2':
                result.append(f'{folder.name}-{suffix}')

        return tuple(sorted(result))

    def brain_mask(i, _base):
        return np.bool_(nb.load(_base / 'METADATA' / 'Brain_mask.nii.gz').get_fdata())

    def deface_mask(i, _base):
        return np.bool_(nb.load(_base / 'METADATA' / 'Deface_mask.nii.gz').get_fdata())

    def _image_file(i, _base):
        i, suffix = i.rsplit('-', 1)
        return nb.load(_base / 'SUBJECTS' / i / f'{suffix}.nii.gz')

    def modality(i):
        _, suffix = i.rsplit('-', 1)
        return suffix

    def subject_id(i):
        subject, _ = i.rsplit('-', 1)
        return subject

    def affine(_image_file):
        return _image_file.affine

    @deprecate(message='Use `spacing` method instead.')
    def voxel_spacing(spacing: Output):
        return spacing

    def spacing(_image_file):
        # voxel spacing is [1, 1, 1] for all images in this dataset...
        return tuple(_image_file.header['pixdim'][1:4])

    def image(_image_file):
        # intensities are not integer-valued in this dataset...
        return np.asarray(_image_file.dataobj)

    def _metadata(i, _base):
        i, _ = i.rsplit('-', 1)
        return load(_base / 'SUBJECTS' / i / 'metadata.json')

    def genetic_and_histological_label_idh(_metadata):
        return _metadata['Genetic_and_Histological_labels']['IDH']

    def genetic_and_histological_label_1p19q(_metadata):
        return _metadata['Genetic_and_Histological_labels']['1p19q']

    def genetic_and_histological_label_grade(_metadata):
        return _metadata['Genetic_and_Histological_labels']['Grade']

    def age(_metadata):
        return _metadata['Clinical_data']['Age']

    def sex(_metadata):
        return _metadata['Clinical_data']['Sex']

    def observer(_metadata):
        return _metadata['Segmentation_source']['Observer']

    def original_scan(_metadata):
        return _metadata['Segmentation_source']['Original scan']

    def manufacturer(_metadata):
        return _metadata['Scan_characteristics']['Manufacturer']

    def system(_metadata):
        return _metadata['Scan_characteristics']['System']

    def field(_metadata):
        return _metadata['Scan_characteristics']['Field']

    def mask(i, _base):
        i, _ = i.rsplit('-', 1)
        return np.bool_(nb.load(_base / 'SUBJECTS' / i / 'MASK.nii.gz').get_fdata())



import contextlib
import datetime
import functools
import zipfile
from gzip import GzipFile
from pathlib import Path
from typing import List, Union

import nibabel
import numpy as np
from dicom_csv import get_common_tag, order_series, stack_images
from dicom_csv.exceptions import ConsistencyError, TagTypeError
from pydicom import Dataset, dcmread


Numeric = Union[float, int]


@contextlib.contextmanager
def unpack(root: str, relative: str, archive_root_name: str = None, archive_ext: str = None):
    """Provides the absolute path to the file in both scenarios: inside archive or inside folder.

    Parameters
    ----------
    root : str, Path
        Absolute path to the downloaded archive or the unpacked archive root.
    relative : str, Path
        Relative file path inside the archive. Archive's root folder sholud be ommited.
    archive_root_name : str, Path, optional
        If `root` is a archive, it's root folder name shold be given.
    archive_ext: {'.zip'}, optional
        Compression algorithm used to create the archive

    Returns
    -------
    unpacked : Path
        Absolute file path to be opened.
    is_unpacked : {True, False}
        Reached file state. `True` if the file is located inside archive, `False` otherwise.
    """
    unpacked = Path(root) / relative

    if unpacked.exists():
        yield unpacked, True
    elif archive_ext == '.zip':
        with zipfile.Path(root, str(Path(archive_root_name, relative))).open('rb') as unpacked:
            yield unpacked, False
    else:
        raise ValueError('Unexpected file path or unsupported compression algorithm.')


@contextlib.contextmanager
def open_nii_gz_file(unpacked):
    """Opens ``.nii.gz`` file if it is packed in archive

    Examples
    --------
    >>> with unpack('/path/to/archive.zip', 'relative/file/path', 'root', '.zip') as (unpacked, is_unpacked):
    >>>     with open_nii_gz_file(unpacked) as image:
    >>>         print(np.asarray(image.dataobj).shape)
    # (512, 512, 256)
    """
    with GzipFile(fileobj=unpacked) as nii:
        nii = nibabel.FileHolder(fileobj=nii)
        yield nibabel.Nifti1Image.from_file_map({'header': nii, 'image': nii})


def get_series_date(series):
    try:
        study_date = get_common_tag(series, 'StudyDate')
    except (TagTypeError, ConsistencyError):
        return

    if not isinstance(study_date, str) or not study_date.isnumeric() or len(study_date) != 8:
        return

    try:
        year = int(study_date[:4])
        month = int(study_date[4:6])
        day = int(study_date[6:])
    except TypeError:
        return

    if year < 1972:  # the year of creation of the first CT scanner
        return

    return datetime.date(year, month, day)


def propagate_none(func):
    @functools.wraps(func)
    def wrapper(x, *args, **kwargs):
        return None if (x is None) else func(x, *args, **kwargs)

    return wrapper


def deprecate(message=None):
    def decorator(func):
        return functools.wraps(func)(np.deprecate(message=message)(func))

    return decorator


def image_from_dicom_folder(folder: Union[str, Path]) -> np.ndarray:
    return stack_images(series_from_dicom_folder(folder))


def series_from_dicom_folder(folder: Union[str, Path]) -> List[Dataset]:
    return order_series([dcmread(p) for p in Path(folder).glob('*.dcm')])



from functools import lru_cache
from pathlib import Path
from zipfile import ZipFile

import nibabel
import numpy as np
import pandas as pd
from connectome import Source, meta
from connectome.interface.nodes import Silent

from ..internals import checksum, licenses, register
from ..utils import open_nii_gz_file, unpack
from .utils import add_labels


ARCHIVE_NAME = 'amos22.zip'
ARCHIVE_ROOT_NAME = 'amos22'


@register(
    body_region='Abdomen',
    license=licenses.CC_BY_40,
    link='https://zenodo.org/record/7262581',
    modality=('CT', 'MRI'),
    raw_data_size='23G',
    prep_data_size='89,5G',
    task='Supervised multi-modality abdominal multi-organ segmentation',
)
@checksum('amos')
class AMOS(Source):
    """
    AMOS provides 500 CT and 100 MRI scans collected from multi-center, multi-vendor, multi-modality, multi-phase,
    multi-disease patients, each with voxel-level annotations of 15 abdominal organs, providing challenging examples
    and test-bed for studying robust segmentation algorithms under diverse targets and scenarios. [1]

    Parameters
    ----------
    root : str, Path, optional
        Absolute path to the root containing the downloaded archive and meta.
        If not provided, the cache is assumed to be already populated.

    Notes
    -----
    Download link: https://zenodo.org/record/7262581/files/amos22.zip

    Examples
    --------
    >>> # Download the archive and meta to any folder and pass the path to the constructor:
    >>> ds = AMOS(root='/path/to/the/downloaded/files')
    >>> print(len(ds.ids))
    # 961
    >>> print(ds.image(ds.ids[0]).shape)
    # (768, 768, 90)
    >>> print(ds.mask(ds.ids[26]).shape)
    # (512, 512, 124)

    References
    ----------
    .. [1] JI YUANFENG. (2022). Amos: A large-scale abdominal multi-organ benchmark for
    versatile medical image segmentation [Data set]. Zenodo. https://doi.org/10.5281/zenodo.7262581
    """

    _root: str = None

    add_labels(locals())

    def _base(_root: Silent):
        return Path(_root)

    @meta
    def ids(_id2split):
        return sorted(_id2split)

    def image(i, _id2split, _base):
        file = f'images{_id2split[i]}/amos_{i}.nii.gz'

        with unpack(_base / ARCHIVE_NAME, file, ARCHIVE_ROOT_NAME, '.zip') as (unpacked, is_unpacked):
            if is_unpacked:
                return np.asarray(nibabel.load(unpacked).dataobj)
            else:
                with open_nii_gz_file(unpacked) as image:
                    return np.asarray(image.dataobj)

    def affine(i, _id2split, _base):
        """The 4x4 matrix that gives the image's spatial orientation"""
        file = f'images{_id2split[i]}/amos_{i}.nii.gz'

        with unpack(_base / ARCHIVE_NAME, file, ARCHIVE_ROOT_NAME, '.zip') as (unpacked, is_unpacked):
            if is_unpacked:
                return nibabel.load(unpacked).affine
            else:
                with open_nii_gz_file(unpacked) as image:
                    return image.affine

    def mask(i, _id2split, _base):
        file = f'labels{_id2split[i]}/amos_{i}.nii.gz'

        try:
            with unpack(_base / ARCHIVE_NAME, file, ARCHIVE_ROOT_NAME, '.zip') as (unpacked, is_unpacked):
                if is_unpacked:
                    return np.asarray(nibabel.load(unpacked).dataobj)
                else:
                    with open_nii_gz_file(unpacked) as image:
                        return np.asarray(image.dataobj)
        except FileNotFoundError:
            return None

    @lru_cache(None)
    def _id2split(_base):
        id2split = {}

        with ZipFile(_base / ARCHIVE_NAME) as zf:
            for x in zf.namelist():
                if (len(x.strip('/').split('/')) == 3) and x.endswith('.nii.gz'):
                    file, split = x.split('/')[-1], x.split('/')[-2][-2:]
                    id_ = file.split('.')[0].split('_')[-1]

                    id2split[id_] = split

        return id2split

    @lru_cache(None)
    def _meta(_base):
        file = 'labeled_data_meta_0000_0599.csv'

        with unpack(_base, file) as (unpacked, _):
            return pd.read_csv(unpacked)



import nibabel
import numpy as np

from ..utils import open_nii_gz_file, unpack
from .const import ANATOMICAL_STRUCTURES, LABELS


ARCHIVE_ROOT = 'Totalsegmentator_dataset'


def add_labels(scope):
    def make_loader(label):
        def loader(i, _meta):
            return _meta[_meta['image_id'] == i][label].item()

        return loader

    for label in LABELS:
        scope[label] = make_loader(label)


def add_masks(scope):
    def make_loader(anatomical_structure):
        def loader(i, _base):
            file = f'{i}/segmentations/{anatomical_structure}.nii.gz'

            with unpack(_base, file, ARCHIVE_ROOT, '.zip') as (unpacked, is_unpacked):
                if is_unpacked:
                    return np.asarray(nibabel.load(unpacked).dataobj)
                else:
                    with open_nii_gz_file(unpacked) as image:
                        return np.asarray(image.dataobj)

        return loader

    for anatomical_structure in ANATOMICAL_STRUCTURES:
        scope[anatomical_structure] = make_loader(anatomical_structure)



from .const import COLUMN2LABEL


def add_labels(scope):
    def make_loader(column):
        def loader(i, _meta):
            # ambigous data in meta
            if int(i) in [500, 600]:
                return None

            return _meta[_meta['amos_id'] == int(i)][column].item()

        return loader

    for column, label in COLUMN2LABEL.items():
        scope[label] = make_loader(column)



import contextlib
import gzip
import re
import zipfile
from pathlib import Path
from zipfile import ZipFile

import nibabel as nb
import numpy as np
from connectome import Output, Source, meta
from connectome.interface.nodes import Silent

from .internals import checksum, licenses, register
from .utils import deprecate


@register(
    body_region=('Chest', 'Abdomen'),
    license=licenses.CC_BYSA_40,
    link='https://www.medseg.ai/database/liver-segments-50-cases',
    modality='CT',
    prep_data_size='1,88G',
    raw_data_size='616M',
    task='Segmentation',
)
@checksum('liver_medseg')
class LiverMedseg(Source):
    """
    LiverMedseg is a public CT segmentation dataset with 50 annotated images.
    Case collection of 50 livers with their segments.
    Images obtained from Decathlon Medical Segmentation competition

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.
    version : str, optional
        the data version. Only has effect if the library was installed from a cloned git repository.

    Notes
    -----
    Download links:
    https://www.medseg.ai/database/liver-segments-50-cases

    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> ds = LiverMedseg(root='/path/to/archives/root')
    >>> print(len(ds.ids))
    # 50
    >>> print(ds.image(ds.ids[0]).shape)
    # (512, 512, 38)

    References
    ----------
    """

    _root: str = None

    @meta
    def ids(_root: Silent):
        if _root is None:
            raise ValueError('Please provide the `root` argument')

        result = set()
        with ZipFile(Path(_root) / 'img.zip') as zf:
            for zipinfo in zf.infolist():
                if zipinfo.is_dir():
                    continue
                file_stem = Path(zipinfo.filename).stem
                result.add('liver_medseg_' + re.findall(r'\d+', file_stem)[0])

        return tuple(sorted(result))

    def _file(i: str, _root: Silent):
        if _root is None:
            raise ValueError('Please provide the `root` argument')

        num_id = i.split('_')[-1]
        return zipfile.Path(Path(_root) / 'img.zip', f'img{num_id}.nii.gz')

    def image(_file) -> np.ndarray:
        with open_nii_gz_file(_file) as nii_file:
            return np.asarray(nii_file.dataobj)

    def affine(_file) -> np.ndarray:
        """The 4x4 matrix that gives the image's spatial orientation."""
        with open_nii_gz_file(_file) as nii_file:
            return nii_file.affine

    @deprecate(message='Use `spacing` method instead.')
    def voxel_spacing(spacing: Output) -> tuple:
        return spacing

    def spacing(_file) -> tuple:
        with open_nii_gz_file(_file) as nii_file:
            return tuple(nii_file.header['pixdim'][1:4])

    def mask(_file) -> np.ndarray:
        path = Path(str(_file).replace('img', 'mask'))
        folder, image = path.parent, path.name
        _file = zipfile.Path(folder, image)
        with open_nii_gz_file(_file) as nii_file:
            return np.asarray(nii_file.dataobj).astype(np.uint8)


# TODO: sync with amid.utils
@contextlib.contextmanager
def open_nii_gz_file(file):
    with file.open('rb') as opened:
        with gzip.GzipFile(fileobj=opened) as nii:
            nii = nb.FileHolder(fileobj=nii)
            yield nb.Nifti1Image.from_file_map({'header': nii, 'image': nii})



import os.path
import warnings
from functools import lru_cache
from pathlib import Path
from typing import Tuple

import mdai
import numpy as np
import pandas as pd
import pydicom
from connectome import Output, Source, meta
from connectome.interface.nodes import Silent
from dicom_csv import (
    drop_duplicated_instances,
    drop_duplicated_slices,
    expand_volumetric,
    get_pixel_spacing,
    get_slice_locations,
    join_tree,
    order_series,
    stack_images,
)
from skimage.draw import polygon

from .internals import checksum, licenses, register


@register(
    body_region='Thorax',
    license=licenses.CC_BYNC_40,
    link='https://wiki.cancerimagingarchive.net/pages/viewpage.action?pageId=80969742',
    modality='CT',
    prep_data_size='7,9G',
    raw_data_size='12G',
    task='COVID-19 Segmentation',
)
@checksum('midrc')
class MIDRC(Source):
    """

        MIDRC-RICORD dataset 1a is a public COVID-19 CT segmentation dataset with 120 scans.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.
    version : str, optional
        the data version. Only has effect if the library was installed from a cloned git repository.

    Notes
    -----
    Follow the download instructions at https://wiki.cancerimagingarchive.net/pages/viewpage.action?pageId=80969742
    Download both Images and Annotations to the same folder

    Then, the folder with downloaded data should contain two paths with the data

    The folder should have this structure:
        <...>/<MIDRC-root>/MIDRC-RICORD-1A
        <...>/<MIDRC-root>/MIDRC-RICORD-1a_annotations_labelgroup_all_2020-Dec-8.json


    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> ds = MIDRC(root='/path/to/downloaded/data/folder/')
    >>> print(len(ds.ids))
     155
    >>> print(ds.image(ds.ids[0]).shape)
     (512, 512, 112)
    >>> print(ds.mask(ds.ids[80]).shape)
     (6, 512, 512, 450)

    References
    ----------
    """

    _root: str = None
    _pathologies: Tuple[str] = (
        'Infectious opacity',
        'Infectious TIB/micronodules',
        'Atelectasis',
        'Other noninfectious opacity',
        'Noninfectious nodule/mass',
        'Infectious cavity',
    )

    def _base(_root: Silent):
        if _root is None:
            raise ValueError('Please provide the `root` argument')
        return Path(_root)

    @lru_cache(None)
    def _joined(_base):
        joined_path = _base / 'joined.csv'
        if joined_path.exists():
            return pd.read_csv(joined_path)
        joined = join_tree(_base / 'MIDRC-RICORD-1A', verbose=1)
        joined = joined[[x.endswith('.dcm') for x in joined.FileName]]
        joined.to_csv(_base / 'joined.csv')
        return joined

    @meta
    def ids(_joined):
        return tuple(_joined['SeriesInstanceUID'].unique())

    def _annotation(_base):
        json_path = 'MIDRC-RICORD-1a_annotations_labelgroup_all_2020-Dec-8.json'
        # TODO: do we really need a whole lib to parse one json??? it also generates annoying pandas warning
        return mdai.common_utils.json_to_dataframe(_base / json_path)['annotations']

    def _series(i, _base, _joined):
        sub = _joined[_joined.SeriesInstanceUID == i]
        series_files = sub['PathToFolder'] + os.path.sep + sub['FileName']
        series_files = [_base / 'MIDRC-RICORD-1A' / x for x in series_files]
        series = list(map(pydicom.dcmread, series_files))
        # series = sorted(series, key=lambda x: x.InstanceNumber)
        series = expand_volumetric(series)
        series = drop_duplicated_instances(series)

        if True:  # drop_dupl_slices
            _original_num_slices = len(series)
            series = drop_duplicated_slices(series)
            if len(series) < _original_num_slices:
                warnings.warn(f'Dropped duplicated slices for series {series[0]["StudyInstanceUID"]}.')

        series = order_series(series, decreasing=False)
        return series

    def image(_series):
        image = stack_images(_series, -1).astype(np.int16).transpose(1, 0, 2)
        return image

    def image_meta(_series):
        metas = [list(dict(s).values()) for s in _series]
        result = {}
        for meta_ in metas:
            for element in meta_:
                if element.keyword in ['PixelData']:
                    continue
                if element.keyword not in result:
                    result[element.keyword] = [element.value]
                elif result[element.keyword][-1] != element.value:
                    result[element.keyword].append(element.value)
        # turn elements that are the same across the series back from array
        result = {k: v[0] if len(v) == 1 else v for k, v in result.items()}
        return result

    def _study_id(i, _joined):
        study_ids = _joined[_joined.SeriesInstanceUID == i].StudyInstanceUID.unique()
        assert len(study_ids) == 1
        # series_id_to_study
        return study_ids[0]

    def spacing(_series):
        pixel_spacing = get_pixel_spacing(_series).tolist()
        slice_locations = get_slice_locations(_series)
        diffs, counts = np.unique(np.round(np.diff(slice_locations), decimals=5), return_counts=True)
        spacing = np.float32([pixel_spacing[1], pixel_spacing[0], diffs[np.argsort(counts)[-1]]])
        return tuple(spacing.tolist())

    def labels(_study_id, _annotation):
        sub = _annotation[(_annotation.scope == 'STUDY') & (_annotation.StudyInstanceUID == _study_id)]
        return tuple(sub['labelName'].unique())

    def mask(i, image_meta: Output, _annotation, _pathologies):
        # TODO: mask has 6 channels now. Consider adding different methods ot at least a docstring naming channels...
        sub = _annotation[(_annotation.SeriesInstanceUID == i) & (_annotation.scope == 'INSTANCE')]
        shape = (image_meta['Rows'], image_meta['Columns'], len(image_meta['SOPInstanceUID']))
        mask = np.zeros((len(_pathologies), *shape), dtype=bool)
        if len(sub) == 0:
            return None
        for label, row in sub.iterrows():
            pathology_index = _pathologies.index(row['labelName'])
            slice_index = image_meta['SOPInstanceUID'].index(row['SOPInstanceUID'])
            if row['data'] is None:
                warnings.warn(f'{label} annotations for series {i} contains None for slice {slice_index}.')
                continue
            ys, xs = np.array(row['data']['vertices']).T
            mask[(pathology_index, *polygon(ys, xs, shape[:2]), slice_index)] = True
        return mask



from pathlib import Path

import typer
from bev.cli.app import command
from bev.cli.init import init as bev_init

from .registry import gather_datasets


# older versions of typer don't have this arg, and it's simpler to avoid conflicts this way
try:
    app = typer.Typer(pretty_exceptions_show_locals=False)
except TypeError:
    app = typer.Typer()


def main():
    app()


@command(app)
def init(
    permissions: str = typer.Option(
        None,
        '--permissions',
        '-p',
        help='The permissions mask used to create the storage, e.g. 770',
    ),
    group: str = typer.Option(
        None,
        '--group',
        '-g',
        help='The group used to create the storage',
    ),
):
    return bev_init(Path(__file__).resolve().parent.parent / 'data', permissions, group)


@app.command()
def populate(
    dataset: str = typer.Argument(..., help='The dataset to populate'),
    root: Path = typer.Argument(..., help='The path to the downloaded raw data'),
    ignore_errors: bool = typer.Option(
        False,
        help='Whether to ignore all the exception during population. '
        'Warning! The failed ids will be excluded from the populated dataset',
    ),
    n_jobs: int = typer.Option(1, help='How many threads to use for population'),
    fetch: bool = typer.Option(False, help='Whether to fetch the missing data from remote locations, if any'),
):
    try:
        cls = gather_datasets()[dataset][0]
        ds = cls(root=root)
        success, errors = ds._populate(n_jobs=n_jobs, fetch=fetch, ignore_errors=ignore_errors)
        print(f'Total added: {success} entries, and encountered {errors} errors')
    except KeyboardInterrupt as e:
        raise RuntimeError from e



from bev import Repository
from bev.exceptions import ConfigError


def get_repo(path=None, strict=True):
    try:
        repo = Repository.from_here('../data')
    except (ConfigError, FileNotFoundError):
        if strict:
            raise
        return

    if path:
        repo = repo / path
    return repo



from contextlib import suppress
from functools import lru_cache
from pathlib import Path

import pandas as pd
import pydicom
from bev.utils import PathOrStr
from connectome import Source, meta
from connectome.interface.nodes import Silent

from ..internals import checksum, register
from .utils import add_csv_fields, fields, unpack


@register(
    body_region='Thorax',
    license='Non-Commercial Use',
    link='https://www.kaggle.com/competitions/rsna-breast-cancer-detection/data',
    modality='MG',
    raw_data_size='271G',
    prep_data_size='294G',
    task='Breast cancer classification',
)
@checksum('rsna-breast-cancer', cache_columns=tuple(fields))
class RSNABreastCancer(Source):
    _root: PathOrStr

    def _base(_root: Silent):
        return Path(_root)

    @lru_cache(None)
    def _meta(_base):
        dfs = []
        for part in 'train', 'test':
            with suppress(FileNotFoundError):
                with unpack(_base, f'{part}.csv') as (file, _):
                    df = pd.read_csv(file)
                    df['part'] = part
                    dfs.append(df)

        if not dfs:
            raise FileNotFoundError('No metadata found')
        dfs = pd.concat(dfs, ignore_index=True)
        for field in 'image_id', 'patient_id', 'site_id':
            dfs[field] = dfs[field].astype(str)

        raw = list(map(str, dfs.image_id.tolist()))
        ids = set(raw)
        if len(ids) != len(raw):
            raise ValueError('The image ids are not unique')

        return {row.image_id: row for _, row in dfs.iterrows()}

    def _row(i, _meta):
        return _meta[i]

    add_csv_fields(locals())

    @meta
    def ids(_meta):
        return tuple(sorted(_meta))

    def _dicom(_row, _base):
        with unpack(_base, f'{_row.part}_images/{_row.patient_id}/{_row.image_id}.dcm') as (file, _):
            return pydicom.dcmread(file)

    def image(_dicom):
        return _dicom.pixel_array

    def padding_value(_dicom):
        return getattr(_dicom, 'PixelPaddingValue', None)

    def intensity_sign(_dicom):
        return getattr(_dicom, 'PixelIntensityRelationshipSign', None)



import contextlib
import gzip
import zipfile
from pathlib import Path
from zipfile import ZipFile

import nibabel as nb
import numpy as np
from connectome import Output, Source, meta
from connectome.interface.nodes import Silent

from ..internals import checksum, licenses, register
from ..utils import deprecate


@register(
    body_region='Head',
    license=licenses.CC_BYND_40,
    link='https://sites.google.com/view/calgary-campinas-dataset/home',
    modality='MRI T1',
    prep_data_size='14,66G',
    raw_data_size='4,1G',
    task='Segmentation',
)
@checksum('cc359')
class CC359(Source):
    """
    A (C)algary-(C)ampinas public brain MR dataset with (359) volumetric images [1]_.

    There are three segmentation tasks on this dataset: (i) brain, (ii) hippocampus, and
    (iii) White-Matter (WM), Gray-Matter (WM), and Cerebrospinal Fluid (CSF) segmentation.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.
    version : str, optional
        the data version. Only has effect if the library was installed from a cloned git repository.

    Notes
    -----
    homepage (upd): https://sites.google.com/view/calgary-campinas-dataset/home
    homepage (old): https://miclab.fee.unicamp.br/calgary-campinas-359-updated-05092017

    To obtain MR images and brain and hippocampus segmentation masks, please, follow the instructions
    at the download platform: https://portal.conp.ca/dataset?id=projects/calgary-campinas.

    Via `datalad` lib you need to download three zip archives:
        - `Original.zip` (the original MR images)
        - `hippocampus_staple.zip` (Silver-standard hippocampus masks generated using STAPLE)
        - `Silver-standard-machine-learning.zip` (Silver-standard brain masks generated using a machine learning method)

    To the current date, WM, GM, and CSF mask could be downloaded only from the google drive:
    https://drive.google.com/drive/u/0/folders/0BxLb0NB2MjVZNm9JY1pWNFp6WTA?resourcekey=0-2sXMr8q-n2Nn6iY3PbBAdA.

    Here you need to manually download a folder (from the google drive root above)
    `CC359/Reconstructed/CC359/WM-GM-CSF/`

    So the root folder to pass to this dataset class should contain four objects:
        - three zip archives (`Original.zip`, `hippocampus_staple.zip`, and `Silver-standard-machine-learning.zip`)
        - one folder `WM-GM-CSF` with the original structure:
            <...>/WM-GM-CSF/CC0319_ge_3_45_M.nii.gz
            <...>/WM-GM-CSF/CC0324_ge_3_56_M.nii.gz
            ...

    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> cc359 = CC359(root='/path/to/downloaded/data/folder/')
    >>> print(len(cc359.ids))
    # 359
    >>> print(cc359.image(cc359.ids[0]).shape)
    # (171, 256, 256)
    >>> print(cc359.wm_gm_csf(cc359.ids[80]).shape)
    # (180, 240, 240)

    References
    ----------
    .. [1] Souza, Roberto, et al. "An open, multi-vendor, multi-field-strength brain MR dataset
           and analysis of publicly available skull stripping methods agreement."
           NeuroImage 170 (2018): 482-494.
           https://www.sciencedirect.com/science/article/pii/S1053811917306687

    """

    _root: str = None

    @meta
    def ids(_root):
        result = set()
        with ZipFile(Path(_root) / 'Original.zip') as zf:
            for zipinfo in zf.infolist():
                if zipinfo.is_dir():
                    continue

                file_name = Path(zipinfo.filename).name
                if file_name.startswith('CC'):
                    result.add(file_name.split('_')[0])

        return tuple(sorted(result))

    def _image_file(i, _root: Silent):
        return get_zipfile(i, 'Original.zip', _root)

    def vendor(_image_file):
        return zipfile2meta(_image_file)['vendor']

    def field(_image_file):
        return zipfile2meta(_image_file)['field']

    def age(_image_file):
        return zipfile2meta(_image_file)['age']

    def gender(_image_file):
        return zipfile2meta(_image_file)['gender']

    def image(_image_file):
        with open_nii_gz_file(_image_file) as nii_image:
            return np.asarray(nii_image.dataobj)

    def affine(_image_file):
        """The 4x4 matrix that gives the image's spatial orientation."""
        with open_nii_gz_file(_image_file) as nii_image:
            return nii_image.affine

    @deprecate(message='Use `spacing` method instead.')
    def voxel_spacing(spacing: Output):
        return spacing

    def spacing(_image_file):
        """Returns voxel spacing along axes (x, y, z)."""
        with open_nii_gz_file(_image_file) as nii_image:
            return tuple(nii_image.header['pixdim'][1:4])

    # masks:

    def brain(i, _root: Silent):
        zf = get_zipfile(i, 'Silver-standard-machine-learning.zip', _root)
        with open_nii_gz_file(zf) as nii_image:
            return np.uint8(nii_image.get_fdata())

    def hippocampus(i, _root: Silent):
        try:
            zf = get_zipfile(i, 'hippocampus_staple.zip', _root)
        except KeyError:
            return None

        with open_nii_gz_file(zf) as nii_image:
            return np.uint8(nii_image.get_fdata())

    def wm_gm_csf(i, _root: Silent):
        for file in (Path(_root) / 'WM-GM-CSF').glob('*'):
            if file.name.startswith(i):
                with open_nii_gz_file(file) as nii_image:
                    return np.uint8(nii_image.get_fdata())


# TODO: sync with amid.utils
@contextlib.contextmanager
def open_nii_gz_file(file):
    with file.open('rb') as opened:
        with gzip.GzipFile(fileobj=opened) as nii:
            nii = nb.FileHolder(fileobj=nii)
            yield nb.Nifti1Image.from_file_map({'header': nii, 'image': nii})


def get_zipfile(_id, archive_name, root):
    archive = Path(root) / archive_name
    with ZipFile(archive) as zf:
        for zipinfo in zf.infolist():
            if Path(zipinfo.filename).name.startswith(_id) and not zipinfo.is_dir():
                return zipfile.Path(str(archive), zipinfo.filename)

    raise KeyError(f'Id "{_id}" not found')


def zipfile2meta(zf):
    return dict(zip(['id', 'vendor', 'field', 'age', 'gender'], zf.name[: -len('.nii.gz')].split('_')))



import zipfile
from pathlib import Path
from zipfile import ZipFile

import nibabel as nb
import numpy as np
from connectome import Output, Source, meta
from connectome.interface.nodes import Silent

from ..internals import checksum, licenses, register
from ..utils import deprecate


@register(
    body_region='Abdominal',
    license=licenses.CC_BYNCND_40,
    link='https://competitions.codalab.org/competitions/17094',
    modality='CT',
    prep_data_size='24,7G',
    raw_data_size='35G',
    task='Segmentation',
)
@checksum('lits')
class LiTS(Source):
    """
    A (Li)ver (T)umor (S)egmentation dataset [1]_ from Medical Segmentation Decathlon [2]_

    There are two segmentation tasks on this dataset: liver and liver tumor segmentation.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.
    version : str, optional
        the data version. Only has effect if the library was installed from a cloned git repository.

    Notes
    -----
    Follow the download instructions at https://competitions.codalab.org/competitions/17094.

    Then, the folder with raw downloaded data should contain two zip archives with the train data
    (`Training_Batch1.zip` and `Training_Batch2.zip`)
    and a folder with the test data
    (`LITS-Challenge-Test-Data`).

    The folder with test data should have original structure:
        <...>/LITS-Challenge-Test-Data/test-volume-0.nii
        <...>/LITS-Challenge-Test-Data/test-volume-1.nii
        ...

    P.S. Organs boxes are also provided from a separate source https://github.com/superxuang/caffe_3d_faster_rcnn.

    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> ds = LiTS(root='/path/to/downloaded/data/folder/')
    >>> print(len(ds.ids))
    # 201
    >>> print(ds.image(ds.ids[0]).shape)
    # (512, 512, 163)
    >>> print(ds.tumor_mask(ds.ids[80]).shape)
    # (512, 512, 771)

    References
    ----------
    .. [1] Bilic, Patrick, et al. "The liver tumor segmentation benchmark (lits)."
           arXiv preprint arXiv:1901.04056 (2019).
    .. [2] Antonelli, Michela, et al. "The medical segmentation decathlon."
           arXiv preprint arXiv:2106.05735 (2021).
    """

    _root: str = None

    @meta
    def ids(_root: Silent):
        result = set()
        # zip archives for train images:
        for archive in Path(_root).glob('*.zip'):
            with ZipFile(archive) as zf:
                for zipinfo in zf.infolist():
                    if zipinfo.is_dir():
                        continue

                    file_stem = Path(zipinfo.filename).stem
                    if 'volume' in file_stem:
                        result.add('lits-train-' + file_stem.split('-')[-1])

        # folder for test images:
        for file in (Path(_root) / 'LITS-Challenge-Test-Data').glob('*'):
            result.add('lits-test-' + file.stem.split('-')[-1])

        return tuple(sorted(result))

    def fold(i, _root: Silent):
        num_id = i.split('-')[-1]

        if 'train' in i:
            for archive in Path(_root).glob('*.zip'):
                batch = '1' if ('1' in archive.stem) else '2'

                with ZipFile(archive) as zf:
                    for zipinfo in zf.infolist():
                        if zipinfo.is_dir():
                            continue

                        if num_id == Path(zipinfo.filename).stem.split('-')[-1]:
                            return f'train_batch_{batch}'

        else:  # if 'test' in i:
            return 'test'

    def _file(i, _root: Silent):
        num_id = i.split('-')[-1]

        if 'train' in i:
            for archive in Path(_root).glob('*.zip'):
                with ZipFile(archive) as zf:
                    for zipinfo in zf.infolist():
                        if zipinfo.is_dir():
                            continue

                        file = Path(zipinfo.filename)
                        if ('volume' in file.stem) and (num_id == file.stem.split('-')[-1]):
                            return zipfile.Path(str(archive), str(file))

        else:  # if 'test' in i:
            return Path(_root) / 'LITS-Challenge-Test-Data' / f'test-volume-{num_id}.nii'

        raise KeyError(f'Id "{i}" not found')

    def image(_file):
        with _file.open('rb') as nii:
            nii = nb.FileHolder(fileobj=nii)
            image = nb.Nifti1Image.from_file_map({'header': nii, 'image': nii})
            # most ct scans are integer-valued, this will help us improve compression rates
            return np.int16(image.get_fdata())

    def affine(_file):
        """The 4x4 matrix that gives the image's spatial orientation."""
        with _file.open('rb') as nii:
            nii = nb.FileHolder(fileobj=nii)
            image = nb.Nifti1Image.from_file_map({'header': nii, 'image': nii})
            return image.affine

    @deprecate(message='Use `spacing` method instead.')
    def voxel_spacing(spacing: Output):
        return spacing

    def spacing(_file):
        """Returns voxel spacing along axes (x, y, z)."""
        with _file.open('rb') as nii:
            nii = nb.FileHolder(fileobj=nii)
            image = nb.Nifti1Image.from_file_map({'header': nii, 'image': nii})
            return tuple(image.header['pixdim'][1:4])

    def mask(_file):
        if 'test' not in _file.name:
            with (_file.parent / _file.name.replace('volume', 'segmentation')).open('rb') as nii:
                nii = nb.FileHolder(fileobj=nii)
                image = nb.Nifti1Image.from_file_map({'header': nii, 'image': nii})
                return np.uint8(image.get_fdata())



import gzip
import zipfile
from pathlib import Path
from zipfile import ZipFile

import nibabel
import numpy as np
from connectome import Source, meta
from connectome.interface.nodes import Silent

from .internals import checksum, register


@register(
    body_region='Abdomen',
    license=None,
    link='https://flare22.grand-challenge.org/',
    modality='CT',
    prep_data_size='347G',
    raw_data_size='247G',
    task='Semi-supervised abdominal organ segmentation',
)
@checksum('flare2022')
class FLARE2022(Source):
    """
    An abdominal organ segmentation dataset for semi-supervised learning [1]_.

    The dataset was used at the MICCAI FLARE 2022 challenge.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.

    Notes
    -----
    Download link: https://flare22.grand-challenge.org/Dataset/

    The `root` folder should contain the two downloaded folders, namely: "Training" and "Validation".

    Examples
    --------
    >>> # Place the downloaded folders in any folder and pass the path to the constructor:
    >>> ds = FLARE2022(root='/path/to/downloaded/data/folder/')
    >>> print(len(ds.ids))
    # 2100
    >>> print(ds.image(ds.ids[0]).shape)
    # (512, 512, 110)
    >>> print(ds.mask(ds.ids[25]).shape)
    # (512, 512, 104)

    References
    ----------
    .. [1] Ma, Jun, et al. "Fast and Low-GPU-memory abdomen CT organ segmentation: The FLARE challenge."
    Medical Image Analysis 82 (2022): 102616.
    """

    _root: str = None

    @meta
    def ids(_root: Silent):
        if _root is None:
            raise ValueError('Please pass the locations of the downloaded data.')

        result = set()

        # 50 Training Labeled cases
        archive = Path(_root) / 'Training' / 'FLARE22_LabeledCase50' / 'images.zip'
        with ZipFile(archive) as zf:
            for file in zf.namelist():
                result.add(f"TL{file.split('_')[-2]}")

        # 2000 Training Unlabeled cases
        for archive in (Path(_root) / 'Training').glob('*.zip'):
            with ZipFile(archive) as zf:
                for file in zf.namelist():
                    if not file.endswith('.nii.gz'):
                        continue

                    file = Path(file)
                    result.add(f"TU{file.name.split('_')[-2]}")

        # 50 Validation Unlabeled cases
        for file in (Path(_root) / 'Validation').glob('*'):
            if not file.name.endswith('.nii.gz'):
                continue

            result.add(f"VU{file.name.split('_')[-2]}")

        return sorted(result)

    def _file(i, _root: Silent):
        # 50 Training Labeled cases
        if i.startswith('TL'):
            archive = Path(_root) / 'Training' / 'FLARE22_LabeledCase50' / 'images.zip'
            with ZipFile(archive) as zf:
                for file in zf.namelist():
                    if i[2:] in file:
                        return zipfile.Path(archive, file)

        # 2000 Training Unlabeled cases
        for archive in (Path(_root) / 'Training').glob('*.zip'):
            with ZipFile(archive) as zf:
                for file in zf.namelist():
                    if i[2:] in file:
                        return zipfile.Path(archive, file)

        # 50 Validation Unlabeled cases
        if i.startswith('VU'):
            file = Path(_root) / 'Validation' / f'FLARETs_{i[2:]}_0000.nii.gz'
            return file

        raise ValueError(f'Id "{i}" not found')

    def image(_file):
        with _file.open('rb') as opened:
            with gzip.GzipFile(fileobj=opened) as nii:
                nii = nibabel.FileHolder(fileobj=nii)
                image = nibabel.Nifti1Image.from_file_map({'header': nii, 'image': nii})
                return np.asarray(image.dataobj)

    def affine(_file):
        """The 4x4 matrix that gives the image's spatial orientation"""
        with _file.open('rb') as opened:
            with gzip.GzipFile(fileobj=opened) as nii:
                nii = nibabel.FileHolder(fileobj=nii)
                image = nibabel.Nifti1Image.from_file_map({'header': nii, 'image': nii})
                return image.affine

    def mask(i, _root: Silent):
        if not i.startswith('TL'):
            return None

        archive = Path(_root) / 'Training' / 'FLARE22_LabeledCase50' / 'labels.zip'
        with ZipFile(archive) as zf:
            for file in zf.namelist():
                if i[2:] in file:
                    with zipfile.Path(archive, file).open('rb') as opened:
                        with gzip.GzipFile(fileobj=opened) as nii:
                            nii = nibabel.FileHolder(fileobj=nii)
                            mask = nibabel.Nifti1Image.from_file_map({'header': nii, 'image': nii})
                            return np.asarray(mask.dataobj)



import os

import numpy as np
import pylidc as pl
from connectome import Source, meta
from connectome.interface.nodes import Output, Silent
from dicom_csv import expand_volumetric, get_common_tag, get_orientation_matrix, get_tag, order_series, stack_images
from pylidc.utils import consensus
from scipy import stats

from ..internals import checksum, licenses, register
from ..utils import deprecate, get_series_date
from .nodules import get_nodule


@register(
    body_region='Chest',
    license=licenses.CC_BY_30,
    link='https://wiki.cancerimagingarchive.net/pages/viewpage.action?pageId=1966254',
    modality='CT',
    prep_data_size='71,2G',
    raw_data_size='126G',
    task='Lung nodules segmentation',
)
@checksum('lidc')
class LIDC(Source):
    """
    The (L)ung (I)mage (D)atabase (C)onsortium image collection (LIDC-IDRI) [1]_
    consists of diagnostic and lung cancer screening thoracic computed tomography (CT) scans
    with marked-up annotated lesions and lung nodules segmentation task.
    Scans contains multiple expert annotations.

    Number of CT scans: 1018.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.
    version : str, optional
        the data version. Only has effect if the library was installed from a cloned git repository.

    Notes
    -----
    Follow the download instructions at https://wiki.cancerimagingarchive.net/pages/viewpage.action?pageId=1966254.

    Then, the folder with raw downloaded data should contain folder `LIDC-IDRI`,
    which contains folders `LIDC-IDRI-*`.

    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> ds = LIDC(root='/path/to/downloaded/data/folder/')
    >>> print(len(ds.ids))
    # 1018
    >>> print(ds.image(ds.ids[0]).shape)
    # (512, 512, 194)
    >>> print(ds.cancer(ds.ids[0]).shape)
    # (512, 512, 194)

    References
    ----------
    .. [1] Armato III, McLennan, et al. "The lung image database consortium (lidc) and image database
    resource initiative (idri): a completed reference database of lung nodules on ct scans."
    Medical physics 38(2) (2011): 915–931.
    https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3041807/
    """

    _root: str = None
    _pylidc_config_start: str = '[dicom]\npath = '

    def _check_config(_root: Silent, _pylidc_config_start):
        if _root is not None:
            if os.path.exists(os.path.expanduser('~/.pylidcrc')):
                with open(os.path.expanduser('~/.pylidcrc'), 'r') as config_file:
                    content = config_file.read()
                if content == f'{_pylidc_config_start}{_root}':
                    return

            # save _root path to ~/.pylidcrc file for pylidc
            with open(os.path.expanduser('~/.pylidcrc'), 'w') as config_file:
                config_file.write(f'{_pylidc_config_start}{_root}')
        return

    @meta
    def ids(_root: Silent, _pylidc_config_start, _check_config):
        result = [scan.series_instance_uid for scan in pl.query(pl.Scan).all()]
        return tuple(sorted(result))

    def _scan(i, _root: Silent, _pylidc_config_start, _check_config):
        _id = i.split('_')[-1]
        return pl.query(pl.Scan).filter(pl.Scan.series_instance_uid == _id).first()

    def _series(_scan):
        series = expand_volumetric(_scan.load_all_dicom_images(verbose=False))
        series = order_series(series)
        return series

    def _shape(_series):
        return stack_images(_series, -1).shape

    def image(_scan):
        return _scan.to_volume(verbose=False)

    def study_uid(_scan):
        return _scan.study_instance_uid

    def series_uid(_scan):
        return _scan.series_instance_uid

    def patient_id(_scan):
        return _scan.patient_id

    def sop_uids(_series):
        return [str(get_tag(i, 'SOPInstanceUID')) for i in _series]

    def pixel_spacing(_scan):
        spacing = _scan.pixel_spacing
        return [spacing, spacing]

    def slice_locations(_scan):
        return _scan.slice_zvals

    @deprecate(message='Use `spacing` method instead.')
    def voxel_spacing(_scan, pixel_spacing: Output):
        """Returns voxel spacing along axes (x, y, z)."""
        spacing = np.float32([pixel_spacing[0], pixel_spacing[0], _scan.slice_spacing])
        return spacing

    def spacing(pixel_spacing: Output, slice_locations: Output):
        """
        Volumetric spacing of the image.
        The maximum relative difference in `slice_locations` < 1e-3
        (except 4 images listed below),
        so we allow ourselves to use the common spacing for the whole 3D image.

        Note
        ----
        The `slice_locations` attribute typically (but not always!) has the constant step.
        In LIDC dataset, only 4 images have difference in `slice_locations` > 1e-3:
            1.3.6.1.4.1.14519.5.2.1.6279.6001.526570782606728516388531252230
            1.3.6.1.4.1.14519.5.2.1.6279.6001.329334252028672866365623335798
            1.3.6.1.4.1.14519.5.2.1.6279.6001.245181799370098278918756923992
            1.3.6.1.4.1.14519.5.2.1.6279.6001.103115201714075993579787468219
        And these differences appear in the maximum of 3 slices.
        Therefore, we consider their impact negligible.
        """
        return (*pixel_spacing, stats.mode(np.diff(slice_locations))[0].item())

    def contrast_used(_scan):
        """If the DICOM file for the scan had any Contrast tag, this is marked as `True`."""
        return _scan.contrast_used

    def is_from_initial(_scan):
        """
        Indicates whether or not this PatientID was tagged as
        part of the initial 399 release.
        """
        return _scan.is_from_initial

    def orientation_matrix(_series):
        return get_orientation_matrix(_series)

    def conv_kernel(_series):
        return get_common_tag(_series, 'ConvolutionKernel', default=None)

    def kvp(_series):
        return get_common_tag(_series, 'KVP', default=None)

    def study_date(_series):
        return get_series_date(_series)

    def accession_number(_series):
        return get_common_tag(_series, 'AccessionNumber', default=None)

    def nodules(_scan):
        nodules = []
        for anns in _scan.cluster_annotations():
            nodule_annotations = []
            for ann in anns:
                nodule_annotations.append(get_nodule(ann))
            nodules.append(nodule_annotations)
        return nodules

    def nodules_masks(_scan):
        nodules = []
        for anns in _scan.cluster_annotations():
            nodule_annotations = []
            for ann in anns:
                nodule_annotations.append(ann.boolean_mask())
            nodules.append(nodule_annotations)
        return nodules

    def cancer(_scan, _shape):
        cancer = np.zeros(_shape, dtype=bool)
        for anns in _scan.cluster_annotations():
            cancer |= consensus(anns, pad=np.inf)[0]

        return cancer



import json
from pathlib import Path

import numpy as np
import pydicom
from connectome import Source, meta
from connectome.interface.nodes import Silent
from dicom_csv import (
    Plane,
    drop_duplicated_slices,
    expand_volumetric,
    get_common_tag,
    get_orientation_matrix,
    get_pixel_spacing,
    get_slice_locations,
    get_slices_plane,
    get_tag,
    order_series,
    stack_images,
)

from .internals import checksum, licenses, register
from .utils import get_series_date


@register(
    body_region='Thorax',
    license=licenses.CC_BY_30,
    link='https://wiki.cancerimagingarchive.net/display/NLST/National+Lung+Screening+Trial',
    modality='CT',
    prep_data_size=None,  # TODO: should be measured...
    raw_data_size=None,  # TODO: should be measured...
    task=None,
)
@checksum('nlst')
class NLST(Source):
    """

        Dataset with low-dose CT scans of 26,254 patients acquired during National Lung Screening Trial.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder (usually called NLST) containing the patient subfolders (like 101426).
        If not provided, the cache is assumed to be already populated.
    version : str, optional
        the data version. Only has effect if the library was installed from a cloned git repository.

    Notes
    -----
    Follow the download instructions at
    https://wiki.cancerimagingarchive.net/display/NLST/National+Lung+Screening+Trial.
    The dicoms should be placed under the following folders' structure:
        <...>/<NLST-root>/<patiend_id>/<study_uid>/<date>/<series_uid>/*.dcm

    Examples
    --------
    >>> ds = NLST(root='/path/to/NLST/')
    >>> print(len(ds.ids))
     ...
    >>> print(ds.image(ds.ids[0]).shape)
     ...
    >>> print(ds.mask(ds.ids[80]).shape)
     ...

    References
    ----------
    """

    _root: str = None

    @meta
    def ids(_root: Silent):
        return tuple(
            path.name
            for path in Path(_root).glob('*/*/*/*')
            if path.is_dir()
            if any(path.iterdir())
            if int(_load_json(path.parent / f'{path.name}.json')['Total'][5]) >= 8  # at least 8 slices
        )

    def _series(i, _root: Silent):
        (folder,) = Path(_root).glob(f'**/{i}')
        series = list(map(pydicom.dcmread, folder.iterdir()))
        series = expand_volumetric(series)
        assert get_common_tag(series, 'Modality') == 'CT'
        assert get_slices_plane(series) == Plane.Axial
        series = drop_duplicated_slices(series)
        series = order_series(series, decreasing=False)
        return series

    def image(_series):
        return np.moveaxis(stack_images(_series, -1).astype(np.int16), 0, 1)

    def study_uid(_series):
        return get_common_tag(_series, 'StudyInstanceUID')

    def series_uid(_series):
        return get_common_tag(_series, 'SeriesInstanceUID')

    def sop_uids(_series):
        return [str(get_tag(i, 'SOPInstanceUID')) for i in _series]

    def pixel_spacing(_series):
        return get_pixel_spacing(_series).tolist()

    def slice_locations(_series):
        return get_slice_locations(_series)

    def orientation_matrix(_series):
        return get_orientation_matrix(_series)

    def conv_kernel(_series):
        return get_common_tag(_series, 'ConvolutionKernel', default=None)

    def kvp(_series):
        return get_common_tag(_series, 'KVP', default=None)

    def patient_id(_series):
        return get_common_tag(_series, 'PatientID', default=None)

    def study_date(_series):
        return get_series_date(_series)

    def accession_number(_series):
        return get_common_tag(_series, 'AccessionNumber', default=None)


def _load_json(file):
    with open(file, 'r') as f:
        return json.load(f)



import io
import shutil
import tempfile
from pathlib import Path
from typing import Any, Callable, Sequence, Tuple, Union
from urllib.error import HTTPError, URLError
from urllib.request import urlretrieve

import requests
from bev.config import RemoteConfig, register
from tarn import RemoteStorage
from tarn.config import HashConfig, StorageConfig
from tarn.digest import key_to_relative
from tarn.utils import Key
from yaml import safe_load


class YandexDisk(RemoteStorage):
    def __init__(self, url: str, optional: bool):
        self.optional = optional
        self.url = url
        self._api_url = 'https://cloud-api.yandex.net/v1/disk/public/resources/'
        self.levels = self.hash = None

    def fetch(
        self, keys: Sequence[Key], store: Callable[[Key, Path], Any], config: HashConfig
    ) -> Sequence[Tuple[Any, bool]]:
        results = []
        with tempfile.TemporaryDirectory() as temp_dir:
            source = Path(temp_dir) / 'source'
            if keys and not self._get_config(config):
                return [(None, False)] * len(keys)

            for key in keys:
                try:
                    self._fetch_tree(key_to_relative(key, self.levels), source)

                    value = store(key, source)
                    shutil.rmtree(source)
                    results.append((value, True))

                except requests.exceptions.ConnectionError:
                    results.append((None, False))

                shutil.rmtree(source, ignore_errors=True)

        return results

    def _get_config(self, reference):
        try:
            if self.levels is None:
                response = requests.get(self._api_url, dict(public_key=self.url, path='/config.yml')).json()
                if 'error' in response:
                    raise requests.exceptions.ConnectionError(f'Error while downloading: {response["error"]}')

                url = response['file']
                with io.StringIO(requests.get(url).text) as stream:
                    config = StorageConfig.parse_obj(safe_load(stream))
                self.hash, self.levels = config.hash, config.levels

            assert self.hash == reference, (self.hash, reference)
            return True

        except requests.exceptions.ConnectionError:
            if not self.optional:
                raise

            return False

    @staticmethod
    def _fetch_file(url, local):
        try:
            urlretrieve(url, str(local))
        except (HTTPError, URLError) as e:
            raise requests.exceptions.ConnectionError from e

    def _fetch_tree(self, relative, local):
        local.mkdir(parents=True, exist_ok=True)

        count, total = 0, 1
        while count < total:
            content = requests.get(
                self._api_url, dict(public_key=self.url, path=str('/' / relative), offset=count)
            ).json()
            if 'error' in content:
                raise requests.exceptions.ConnectionError(f'Error while downloading: {content["error"]}')

            content = content['_embedded']
            total = content['total']
            count += len(content['items'])

            for entry in content['items']:
                name = entry['name']
                if 'file' in entry:
                    self._fetch_file(entry['file'], local / name)
                else:
                    self._fetch_tree(relative / name, local / name)


@register('ya-disk')
class YandexDiskConfig(RemoteConfig):
    url: str

    @classmethod
    def from_string(cls, v):
        return cls(url=v)

    def build(self, root: Union[Path, None], optional: bool) -> Union[RemoteStorage, None]:
        return YandexDisk(self.url, optional)



import gzip
from pathlib import Path

import nibabel
import numpy as np
from connectome import Source, meta
from connectome.interface.nodes import Silent

from .internals import checksum, register


@register(
    body_region='Thorax',
    modality='CT',
    task='COVID-19 Segmentation',
    link='https://mosmed.ai/en/datasets/covid191110/',
    raw_data_size='21G',
)
@checksum('covid_1110')
class MoscowCovid1110(Source):
    """
    The Moscow Radiology COVID-19 dataset.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded files.
        If not provided, the cache is assumed to be already populated.
    version : str, optional
        the data version. Only has effect if the library was installed from a cloned git repository.

    Notes
    -----
    Download links:
    https://mosmed.ai/en/datasets/covid191110/

    Examples
    --------
    >>> # Place the downloaded files in any folder and pass the path to the constructor:
    >>> ds = MoscowCovid1110(root='/path/to/files/root')
    >>> print(len(ds.ids))
    # 1110
    >>> print(ds.image(ds.ids[0]).shape)
    # (512, 512, 43)
    """

    _root: str = None

    @meta
    def ids(_root: Silent):
        if _root is None:
            raise ValueError('Please pass the locations of the zip archives')

        return sorted({f.name[:-7] for f in Path(_root).glob('CT-*/*')})

    def _file(i, _root: Silent):
        return next(Path(_root).glob(f'CT-*/{i}.nii.gz'))

    def image(_file):
        with _file.open('rb') as opened:
            with gzip.GzipFile(fileobj=opened) as nii:
                nii = nibabel.FileHolder(fileobj=nii)
                image = nibabel.Nifti1Image.from_file_map({'header': nii, 'image': nii})
                # most ct scans are integer-valued, this will help us improve compression rates
                #  (instead of using `image.get_fdata()`)
                return np.asarray(image.dataobj)

    def affine(_file):
        with _file.open('rb') as opened:
            with gzip.GzipFile(fileobj=opened) as nii:
                nii = nibabel.FileHolder(fileobj=nii)
                image = nibabel.Nifti1Image.from_file_map({'header': nii, 'image': nii})
                return image.affine

    def label(_file):
        return _file.parent.name[3:]

    def mask(i, _root: Silent):
        path = Path(_root) / 'masks' / f'{i}_mask.nii.gz'
        if not path.exists():
            return

        with path.open('rb') as opened:
            with gzip.GzipFile(fileobj=opened) as nii:
                nii = nibabel.FileHolder(fileobj=nii)
                image = nibabel.Nifti1Image.from_file_map({'header': nii, 'image': nii})
                return np.asarray(image.dataobj) > 0.5



import functools
import tempfile
from collections import defaultdict
from contextlib import suppress
from pathlib import Path

from bev import Local, Repository
from bev.exceptions import HashNotFound
from bev.hash import to_hash
from bev.ops import save_hash
from connectome import Chain
from connectome.cache import Cache
from connectome.containers import EdgesBag, IdentityContext
from connectome.engine import Command, ConstantEdge, Details, Node, StaticGraph, StaticHash, TreeNode
from connectome.layers.cache import CacheToStorage
from connectome.utils import AntiSet, node_to_dict
from joblib import Parallel, delayed
from more_itertools import zip_equal
from tarn import DeserializationError, ReadError
from tqdm.auto import tqdm

from .base import get_repo
from .cache import CacheColumns, CacheToDisk, default_serializer


# TODO: add possibility to check the entire tree without the need to pull anything from remote


def checksum(path: str, *, ignore=(), cache_columns=()):
    def make_cache():
        yield CacheToDisk(AntiSet(('id', *cache_columns)), serializer=serializer)
        if cache_columns:
            yield CacheColumns(cache_columns, serializer=serializer, verbose=True)

    serializer = default_serializer(None)

    def decorator(cls):
        class Checked(Chain):
            def __init__(self, root: str = None, version: str = Local):
                ds = cls(root=root)
                args = [ds]

                if version is not None:
                    repository = get_repo(strict=False)
                    if repository is not None:
                        args.append(
                            CacheAndCheck(
                                set(dir(ds)) - {'id', 'ids', *ignore},
                                repository,
                                path,
                                fetch=True,
                                serializer=serializer,
                                version=version,
                            )
                        )
                        # if repository.cache is not None and repository.cache.local:
                        #     args.extend(make_cache())

                self._version = version
                super().__init__(*args)

            def _populate(
                self, *, ignore_errors: bool = False, cache: bool = True, fetch: bool = True, n_jobs: int = 1
            ):
                repository = get_repo()
                ds = self[0]
                fields = sorted(set(dir(ds)) - {'ids', 'id', *ignore})

                if cache:
                    ds = Chain(ds, *make_cache())
                ids = ds.ids

                ds = ds >> CacheAndCheck(
                    fields,
                    repository,
                    path,
                    fetch=fetch,
                    serializer=serializer,
                    version=self._version,
                    return_tree=True,
                )
                _loader = ds._compile(fields)

                def loader(key):
                    try:
                        return key, _loader(key)
                    except Exception as e:
                        if not ignore_errors:
                            raise RuntimeError(f'Error while processing id {key}') from e

                        return key, None

                print(f'Populating the cache with {len(fields)} fields for {len(ids)} ids')

                checksums = {}
                successes = errors = 0
                with ProgressParallel(n_jobs=n_jobs, backend='threading', tqdm_kwargs=dict(total=len(ids))) as bar:
                    for i, trees in tqdm(bar(map(delayed(loader), ids)), 'Saving the checksums'):
                        if trees is None:
                            errors += 1
                            continue

                        successes += 1
                        for name, tree in zip_equal(fields, trees):
                            for k, v in tree.items():
                                checksums['/'.join((name, i, k))] = v

                save_hash(checksums, to_hash(Path(repository.path / path)), repository.storage)
                return successes, errors

        functools.update_wrapper(Checked, cls, updated=())
        return Checked

    return decorator


class CacheAndCheck(CacheToStorage):
    def __init__(
        self,
        names,
        repository: Repository,
        path,
        *,
        serializer=None,
        return_tree: bool = False,
        fetch: bool = True,
        version,
    ):
        super().__init__(names, False)
        serializer = default_serializer(serializer)
        # name -> identifier -> tree
        checksums = defaultdict(lambda: defaultdict(dict))

        with suppress(HashNotFound, ReadError):
            for key, value in repository.load_tree(to_hash(Path(path)), version=version, fetch=fetch).items():
                name, identifier, relative = key.split('/', 2)
                if name in names:
                    checksums[name][identifier][relative] = value

        self.checksums = dict(checksums)
        self.return_tree = return_tree
        self.storage = repository.storage
        self.serializer = serializer
        self.keys = 'ids'

    def _get_storage(self) -> Cache:
        return self.storage

    def _prepare_container(self, previous: EdgesBag) -> EdgesBag:
        if len(previous.inputs) != 1:
            raise ValueError('The input layer must contain exactly one input')
        details = Details(type(self))
        key = Node(previous.inputs[0].name, details)

        mapping = TreeNode.from_edges(previous.edges)
        forward_outputs = node_to_dict(previous.outputs)

        inputs, outputs, edges = [key], [], []
        # do we need to add a `keys` output:
        if self.checksums or self.keys not in forward_outputs:
            forward_outputs.pop(self.keys, None)
            keys = Node(self.keys, details)
            outputs.append(keys)
            # TODO: what if the ids differ in checksum?
            edges.append(ConstantEdge(tuple(sorted(list(self.checksums.values())[0]))).bind((), keys))

        for name in forward_outputs:
            if name in self.names:
                if not self.impure:
                    self._detect_impure(mapping[forward_outputs[name]], name)

                inp, out = Node(name, details), Node(name, details)
                inputs.append(inp)
                outputs.append(out)
                edges.append(
                    CheckSumEdge(
                        self.checksums.get(name, {}),
                        self.serializer,
                        self._get_storage(),
                        self.return_tree,
                        False,
                    ).bind([inp, key], out)
                )

        return EdgesBag(
            inputs,
            outputs,
            edges,
            IdentityContext(),
            persistent=None,
            virtual=AntiSet(node_to_dict(outputs)),
            optional=None,
        )


class CheckSumEdge(StaticGraph, StaticHash):
    def __init__(self, tree, serializer, storage, return_tree: bool, check: bool):
        super().__init__(arity=2)
        self.return_tree = return_tree
        self._serializer, self._storage = serializer, storage
        self.tree = tree
        self.check = check

    def _make_hash(self, inputs):
        if self.return_tree:
            raise ValueError('In `return tree` mode graph hashing is not supported')
        return inputs[0]

    def evaluate(self):
        identifier = yield Command.ParentValue, 1
        expected = self.tree.get(identifier)

        # 1. try to deserialize directly from the storage
        if expected is not None:
            if self.return_tree:
                return expected

            value, exists = self._deserialize(expected)
            if exists:
                return value

        # 2. noting found - compute the value and save it in the storage
        value = yield Command.ParentValue, 0

        # check consistency
        if (expected is not None and self.check) or self.return_tree:
            # TODO: `check` is not the same as save to storage
            tree = self._serialize(value)
            if expected is not None and self.check and tree != expected:
                raise ValueError(f'Checksum failed for {identifier}. Actual: {tree}, expected: {expected}')
            if self.return_tree:
                return tree

        return value

    def _deserialize(self, tree):
        with tempfile.TemporaryDirectory() as base:
            base = Path(base)
            for k, v in tree.items():
                k = base / k
                k.parent.mkdir(parents=True, exist_ok=True)
                with open(k, 'w') as file:
                    file.write(v)

            try:
                return self._serializer.load(base, self._storage), True
            except ReadError as e:
                if isinstance(e, DeserializationError):
                    locations = {}
                    for k, v in tree.items():
                        try:
                            locations[k] = self._storage.read(lambda x: x, v)
                        except ReadError:
                            pass

                    raise DeserializationError(f'{tree}: {locations}')
                return None, False

    def _serialize(self, value):
        with tempfile.TemporaryDirectory() as base:
            base = Path(base)
            self._serializer.save(value, base)
            tree = {}
            # TODO: this is basically `mirror to storage`
            for file in base.glob('**/*'):
                if file.is_dir():
                    continue

                tree[str(file.relative_to(base))] = self._storage.write(file).hex()

            return tree


# source: https://stackoverflow.com/a/61027781
class ProgressParallel(Parallel):
    def __init__(self, *args, tqdm_kwargs=None, **kwargs):
        self._tqdm_kwargs = tqdm_kwargs or {}
        super().__init__(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        with tqdm(**self._tqdm_kwargs) as self._pbar:
            return super().__call__(*args, **kwargs)

    def print_progress(self):
        if 'total' not in self._tqdm_kwargs:
            self._pbar.total = self.n_dispatched_tasks
        self._pbar.n = self.n_completed_tasks
        self._pbar.refresh()

    def __getattr__(self, name):
        return getattr(self._pbar, name)



import plistlib
import warnings
from ast import literal_eval
from enum import IntEnum
from functools import lru_cache
from pathlib import Path
from typing import NamedTuple

import numpy as np
import pandas as pd
import pydicom
from connectome import Source, Transform, meta
from connectome.interface.nodes import Silent
from dicom_csv import (
    drop_duplicated_instances,
    drop_duplicated_slices,
    expand_volumetric,
    get_orientation_matrix,
    get_pixel_spacing,
    get_slice_locations,
    order_series,
    stack_images,
)
from skimage.draw import polygon

from .internals import checksum, licenses, register


class CoCaClasses(IntEnum):
    LAD = 1
    LCX = 2
    RCA = 3
    LCA = 4


class Calcification(NamedTuple):
    label: str
    contour_px: np.ndarray
    contour_mm: np.ndarray
    center: np.ndarray
    area: float
    hu_min: float
    hu_max: float
    hu_dev: float
    hu_mean: float
    length: float
    total: int


@register(
    body_region=('Coronary', 'Chest'),
    license=licenses.StanfordDSResearch,
    link='https://stanfordaimi.azurewebsites.net/datasets/e8ca74dc-8dd4-4340-815a-60b41f6cb2aa',
    modality='CT',
    prep_data_size=None,  # TODO: should be measured...
    raw_data_size='28G',
    task='Coronary Calcium Segmentation',
)
@checksum('stanford_coca')
class StanfordCoCa(Source):
    """
    A Stanford AIMI's Co(ronary) Ca(lcium) dataset.


    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.
    version : str, optional
        the data version. Only has effect if the library was installed from a cloned git repository.

    Notes
    -----
    Follow the download instructions at
    https://stanfordaimi.azurewebsites.net/datasets/e8ca74dc-8dd4-4340-815a-60b41f6cb2aa.
    You'll need to register and accept the terms of use. After that, copy the files from Azure:

    azcopy copy 'some-generated-access-link' /path/to/downloaded/data/ --recursive=true

    Then, the folder with raw downloaded data should contain two subfolders - a subset with gated coronary CT scans
    and corresponding coronary calcium segmentation masks (`Gated_release_final`)
    and a folder with the non-gated CT scans with corresponding coronary with coronary artery calcium scores
    (`deidentified_nongated`).

    The folder with gated data should have original structure:
        ./Gated_release_final/patient/0/folder-with-dcms/
        ./Gated_release_final/calcium_xml/0.xml
        ...

    The folder with nongated data should have original structure:
        ./deidentified_nongated/0/folder-with-dcms/
        ...

    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> ds = StanfordCoCa(root='/path/to/downloaded/data/folder/')
    >>> print(len(ds.ids))
    # 971
    >>> print(ds.image(ds.ids[0]).shape)
    # (512, 512, 57)

    """

    _root: str = None

    def _split(i):
        return i.split('-')[0]

    def _i(i):
        return i.split('-')[1]

    def _folder_with_images(_split):
        if _split == 'gated':
            return Path('Gated_release_final') / 'patient'
        if _split == 'nongated':
            return 'deidentified_nongated'
        raise ValueError("Unknown split. Use 'gated' or 'nongated' options.")

    def _folder_with_annotations(_split):
        if _split == 'gated':
            return Path('Gated_release_final') / 'calcium_xml'
        if _split == 'nongated':
            return None
        raise ValueError("Unknown split. Use 'gated' or 'nongated' options.")

    @meta
    def ids(_root: Silent):
        gated_ids = tuple(
            sorted('gated-' + x.name for x in (Path(_root) / 'Gated_release_final' / 'patient').iterdir() if x.is_dir())
        )
        nongated_ids = tuple(
            sorted('nongated-' + x.name for x in (Path(_root) / 'deidentified_nongated').iterdir() if x.is_dir())
        )

        return gated_ids + nongated_ids

    def _series(_i, _root: Silent, _folder_with_images):
        folder_with_dicoms = Path(_root) / _folder_with_images / _i
        series = list(map(pydicom.dcmread, folder_with_dicoms.glob('*/*.dcm')))
        # series = sorted(series, key=lambda x: x.InstanceNumber)
        series = expand_volumetric(series)
        series = drop_duplicated_instances(series)

        original_num_slices = len(series)
        series = drop_duplicated_slices(series)
        if len(series) < original_num_slices:
            warnings.warn(f'Dropped duplicated slices for series {series[0]["StudyInstanceUID"]}.')

        series = order_series(series, decreasing=False)

        return series

    def image(i, _series):
        image = stack_images(_series, -1).transpose((1, 0, 2)).astype(np.int16)
        return image

    def _image_meta(_series):
        metas = [list(dict(s).values()) for s in _series]
        result = {}
        for meta_ in metas:
            for element in meta_:
                if element.keyword in ['PixelData']:
                    continue
                if element.keyword not in result:
                    result[element.keyword] = [element.value]
                elif result[element.keyword][-1] != element.value:
                    result[element.keyword].append(element.value)
        # turn elements that are the same across the series back from array
        result = {k: v[0] if len(v) == 1 else v for k, v in result.items()}
        return result

    def series_uid(_image_meta):
        return _image_meta.get('SeriesInstanceUID', None)

    def study_uid(_image_meta):
        return _image_meta.get('StudyInstanceUID', None)

    def pixel_spacing(_series):
        return get_pixel_spacing(_series).tolist() if _series else None

    def slice_locations(_series):
        return get_slice_locations(_series) if _series else None

    def orientation_matrix(_series):
        return get_orientation_matrix(_series) if _series else None

    def _raw_annotations(_i, _root: Silent, _folder_with_annotations):
        """Annotation as it is in xml"""
        if _folder_with_annotations is None:
            warnings.warn("The used split doesn't contain segmentation masks.")
            return None

        try:
            with open(Path(_root) / _folder_with_annotations / f'{_i}.xml', 'rb') as fp:
                annotation = plistlib.load(fp)
                image_annotations = annotation['Images']

        except FileNotFoundError:
            warnings.warn(f'Missing annotation for id: {_i}')
            return None

        return image_annotations

    def calcifications(_raw_annotations):
        """Returns list of Calcifications"""
        if _raw_annotations is None:
            return None

        cacs = []
        for slice_annotation in _raw_annotations:
            for roi in slice_annotation['ROIs']:
                if roi['Area'] > 0:
                    contour_px = np.array([literal_eval(x) for x in roi['Point_px']])
                    contour_mm = np.array([literal_eval(x) for x in roi['Point_mm']])
                    name, center, area = roi['Name'], roi['Center'], roi['Area']
                    hu_min, hu_max, hu_dev, hu_mean = roi['Min'], roi['Max'], roi['Dev'], roi['Mean']
                    total, length = roi['Total'], roi['Length']
                    cacs.append(
                        Calcification(
                            name, contour_px, contour_mm, center, area, hu_min, hu_max, hu_dev, hu_mean, total, length
                        )
                    )
        return cacs

    @lru_cache(None)
    def _scores(_root: Silent, _folder_with_images):
        p = Path(_root) / _folder_with_images / 'scores.xlsx'

        if not p.exists():
            return None

        return pd.read_excel(p, index_col=0)

    def score(_i, _scores):
        if _scores is None:
            return None
        try:
            return _scores.loc[_i + 'A'].to_dict()
        except KeyError:
            warnings.warn(f'Missing scores for idx "{_i}"')
            return None


class ContoursToMask(Transform):
    """Our implementation of transform for the contours. One can implement own logic based on this class"""

    __inherit__ = True
    _throw: bool = False
    _class_abbr: dict = {
        'Left Anterior Descending Artery': 'LAD',
        'Left Circumflex Artery': 'LCA',
        'Right Coronary Artery': 'RCA',
        'Left Coronary Artery': 'LCA',
    }

    def mask(id, calcifications, image, slice_locations, _class_abbr, _throw):
        if calcifications is None:
            return None

        shape = image.shape
        sl = slice_locations
        multiclass_mask = np.zeros(shape, np.uint8)
        try:
            for cac in calcifications:
                assert cac.label in _class_abbr, f'Unexpected class: {cac.label}'
                class_name = _class_abbr[cac.label]
                class_id = CoCaClasses[class_name].value
                slice_location = cac.contour_mm[0][-1]
                slice_id = np.argwhere(sl == slice_location).squeeze()
                assert slice_id, 'Slice where calcification is located is not presented in the series.'
                roi_contour = cac.contour_px
                slice_mask = np.zeros(shape[:2])
                xs, ys = polygon(*roi_contour.T, shape[:2])
                slice_mask[xs, ys] = True
                multiclass_mask[..., slice_id] = (class_id * slice_mask.astype(int)).astype(np.uint8)

        except AssertionError as e:
            if _throw:
                raise e
            else:
                warnings.warn(f"Mask preparation for idx {id} failed with: '{str(e)}'. Returning None")
                return None

        return multiclass_mask



from pathlib import Path

import nibabel as nb
import numpy as np
import pandas as pd
from connectome import Output, Source, meta
from connectome.interface.nodes import Silent

from .internals import checksum, licenses, register
from .utils import deprecate


@register(
    body_region='Head',
    license=licenses.PhysioNet_RHD_150,
    link='https://physionet.org/content/ct-ich/1.3.1/',
    modality='CT',
    prep_data_size='661M',
    raw_data_size='2,8G',
    task='Intracranial hemorrhage segmentation',
)
@checksum('ct_ich')
class CT_ICH(Source):
    """
    (C)omputed (T)omography Images for (I)ntracranial (H)emorrhage Detection and (S)egmentation.

    This dataset contains 75 head CT scans including 36 scans for patients diagnosed with
    intracranial hemorrhage with the following types:
    Intraventricular, Intraparenchymal, Subarachnoid, Epidural and Subdural.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.
    version : str, optional
        the data version. Only has effect if the library was installed from a cloned git repository.

    Notes
    -----
    Data can be downloaded here: https://physionet.org/content/ct-ich/1.3.1/.
    Then, the folder with raw downloaded data should contain folders `ct_scans` and `masks` along with other files.

    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> ds = CT_ICH(root='/path/to/downloaded/data/folder/')
    >>> print(len(ds.ids))
    # 75
    >>> print(ds.image(ds.ids[0]).shape)
    # (512, 512, 39)
    >>> print(ds.mask(ds.ids[0]).shape)
    # (512, 512, 39)
    """

    _root: str = None

    @meta
    def ids(_root: Silent):
        result = [f'ct_ich_{uid:0=3d}' for uid in np.concatenate([range(49, 59), range(66, 131)])]
        return tuple(sorted(result))

    def _image_file(i, _root: Silent):
        num_id = i.split('_')[-1]
        return nb.load(Path(_root) / 'ct_scans' / f'{num_id}.nii')

    def image(_image_file):
        # most CT/MRI scans are integer-valued, this will help us improve compression rates
        return np.int16(_image_file.get_fdata()[...])

    def mask(i, _root: Silent):
        num_id = i.split('_')[-1]
        mask_path = Path(_root) / 'masks' / f'{num_id}.nii'
        ct_scan_nifti = nb.load(mask_path)
        return np.bool_(ct_scan_nifti.get_fdata()[...])

    def affine(_image_file):
        """The 4x4 matrix that gives the image's spatial orientation."""
        return _image_file.affine

    @deprecate(message='Use `spacing` method instead.')
    def voxel_spacing(spacing: Output):
        return spacing

    def spacing(_image_file):
        """Returns voxel spacing along axes (x, y, z)."""
        return tuple(_image_file.header['pixdim'][1:4])

    def _patient_metadata(_root: Silent):
        return pd.read_csv(Path(_root) / 'Patient_demographics.csv', index_col='Patient Number')

    def _diagnosis_metadata(_root: Silent):
        return pd.read_csv(Path(_root) / 'hemorrhage_diagnosis_raw_ct.csv')

    def age(i, _patient_metadata):
        num_id = int(i.split('_')[-1])
        _patient_metadata['Age\n(years)'].loc[num_id]

    def gender(i, _patient_metadata):
        num_id = int(i.split('_')[-1])
        _patient_metadata['Gender'].loc[num_id]

    def intraventricular_hemorrhage(i, _patient_metadata):
        """Returns True if hemorrhage exists and its type is intraventricular."""
        num_id = int(i.split('_')[-1])
        return str(_patient_metadata['Hemorrhage type based on the radiologists diagnosis '].loc[num_id]) != 'nan'

    def intraparenchymal_hemorrhage(i, _patient_metadata):
        """Returns True if hemorrhage was diagnosed and its type is intraparenchymal."""
        num_id = int(i.split('_')[-1])
        return str(_patient_metadata['Unnamed: 4'].loc[num_id]) != 'nan'

    def subarachnoid_hemorrhage(i, _patient_metadata):
        """Returns True if hemorrhage was diagnosed and its type is subarachnoid."""
        num_id = int(i.split('_')[-1])
        return str(_patient_metadata['Unnamed: 5'].loc[num_id]) != 'nan'

    def epidural_hemorrhage(i, _patient_metadata):
        """Returns True if hemorrhage was diagnosed and its type is epidural."""
        num_id = int(i.split('_')[-1])
        return str(_patient_metadata['Unnamed: 6'].loc[num_id]) != 'nan'

    def subdural_hemorrhage(i, _patient_metadata):
        """Returns True if hemorrhage was diagnosed and its type is subdural."""
        num_id = int(i.split('_')[-1])
        return str(_patient_metadata['Unnamed: 7'].loc[num_id]) != 'nan'

    def fracture(i, _patient_metadata):
        """Returns True if skull fracture was diagnosed."""
        num_id = int(i.split('_')[-1])
        return str(_patient_metadata['Fracture (yes 1/no 0)'].loc[num_id]) != 'nan'

    def notes(i, _patient_metadata):
        """Returns special notes if they exist."""
        num_id = int(i.split('_')[-1])
        result = str(_patient_metadata['Note1'].loc[num_id])
        return result if result != 'nan' else None

    def hemorrhage_diagnosis_raw_metadata(i, _diagnosis_metadata):
        num_id = int(i.split('_')[-1])
        return _diagnosis_metadata[_diagnosis_metadata['PatientNumber'] == num_id]



import contextlib
import gzip
import zipfile
from pathlib import Path
from zipfile import ZipFile

import nibabel as nb
import numpy as np
from connectome import Output, Source, meta
from connectome.interface.nodes import Silent

from .internals import checksum, register
from .utils import deprecate


@register(
    body_region=('Head', 'Abdominal'),
    license=None,  # FIXME: inherit licenses from the original datasets...
    link='http://medicalood.dkfz.de/web/',
    modality=('MRI', 'CT'),
    prep_data_size='405G',
    raw_data_size='120G',
    task='Out-of-distribution detection',
)
@checksum('mood')
class MOOD(Source):
    """
    A (M)edival (O)ut-(O)f-(D)istribution analysis challenge [1]_

    This dataset contains raw brain MRI and abdominal CT images.

    Number of training samples:
    - Brain: 800 scans ( 256 x 256 x 256 )
    - Abdominal: 550 scans ( 512 x 512 x 512 )

    For each setup there are 4 toy test samples with OOD cases.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.
    version : str, optional
        the data version. Only has effect if the library was installed from a cloned git repository.

    Notes
    -----
    Follow the download instructions at https://www.synapse.org/#!Synapse:syn21343101/wiki/599515.

    Then, the folder with raw downloaded data should contain four zip archives with data
    (`abdom_toy.zip`, `abdom_train.zip`, `brain_toy.zip` and `brain_train.zip`).

    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> ds = MOOD(root='/path/to/downloaded/data/folder/')
    >>> print(len(ds.ids))
    # 1358
    >>> print(ds.image(ds.ids[0]).shape)
    # (512, 512, 512)
    >>> print(ds.pixel_label(ds.ids[0]).shape)
    # (512, 512, 512)

    References
    ----------
    .. [1] Zimmerer, Petersen, et al. "Medical Out-of-Distribution Analysis Challenge 2022."
           doi: 10.5281/zenodo.6362313 (2022).
    """

    _root: str = None

    @meta
    def ids(_root: Silent):
        result = set()
        # zip archives for train images:
        for archive in Path(_root).glob('*.zip'):
            if 'brain' in str(archive):  # define whether it is brain (MRI) or abdominal (CT)
                task = 'brain'
            else:
                task = 'abdom'

            if 'toy' in str(archive):  # fold - train or toy test
                fold = 'toy'
            else:
                fold = 'train'

            with ZipFile(archive) as zf:
                for zipinfo in zf.infolist():
                    if zipinfo.is_dir():
                        continue

                    file_stem = Path(zipinfo.filename).stem
                    if '.nii' in file_stem:
                        if fold == 'train':
                            result.add(f'mood_{task}_{fold}_{file_stem.split(".nii")[0]}')
                        # fold == 'toy'
                        else:
                            result.add(f'mood_{task}_{file_stem.split(".nii")[0]}')

        return tuple(sorted(result))

    def fold(i, _root: Silent):
        """Returns fold: train or toy (test)."""
        if 'train' in i:
            return 'train'
        # if 'toy' in i:
        return 'toy'

    def task(i, _root: Silent):
        """Returns task: brain (MRI) or abdominal (CT)."""
        if 'brain' in i:
            return 'brain'
        # if 'abdom' in i:
        return 'abdom'

    def _file(i, _root: Silent):
        task, fold, num_id = i.split('_')[-3:]
        if fold == 'train':
            return zipfile.Path(Path(_root) / f'{task}_{fold}.zip', f'{task}_{fold}/{num_id}.nii.gz')
        return zipfile.Path(Path(_root) / f'{task}_{fold}.zip', f'toy/toy_{num_id}.nii.gz')

    def image(_file):
        with open_nii_gz_file(_file) as nii_image:
            return np.asarray(nii_image.dataobj)

    def affine(_file):
        """The 4x4 matrix that gives the image's spatial orientation."""
        with open_nii_gz_file(_file) as nii_image:
            return nii_image.affine

    @deprecate(message='Use `spacing` method instead.')
    def voxel_spacing(spacing: Output):
        return spacing

    def spacing(_file):
        """Returns voxel spacing along axes (x, y, z)."""
        with open_nii_gz_file(_file) as nii_image:
            return tuple(nii_image.header['pixdim'][1:4])

    def sample_label(_file):
        """
        Returns sample-level OOD score for toy examples and None otherwise.
        0 indicates no abnormality and 1 indicates abnormal input.
        """
        if 'toy' in _file.name:
            with (_file.parent.parent / 'toy_label/sample' / f'{_file.name}.txt').open('r') as nii:
                return int(nii.read())

    def pixel_label(_file):
        """
        Returns voxel-level OOD scores for toy examples and None otherwise.
        0 indicates no abnormality and 1 indicates abnormal input.
        """
        if 'toy' in _file.name:
            with open_nii_gz_file(_file.parent.parent / 'toy_label/pixel' / _file.name) as nii_image:
                return np.bool_(nii_image.get_fdata())


# TODO: sync with amid.utils
@contextlib.contextmanager
def open_nii_gz_file(file):
    with file.open('rb') as opened:
        with gzip.GzipFile(fileobj=opened) as nii:
            nii = nb.FileHolder(fileobj=nii)
            yield nb.Nifti1Image.from_file_map({'header': nii, 'image': nii})



import contextlib
import gzip
import tarfile
import typing as tp
from functools import lru_cache
from pathlib import Path

import nibabel as nb
import numpy as np
import pandas as pd
import pydicom
from connectome import Source, Transform, meta
from connectome.interface.nodes import Silent
from deli import load

from .internals import checksum, licenses, register


@register(
    body_region='Chest',
    license=licenses.CC_BY_40,
    link='https://ieee-dataport.org/open-access/bimcv-covid-19'
    '-large-annotated-dataset-rx-and-ct-images-covid-19-patients-0',
    modality='CT',
    prep_data_size='859G',
    raw_data_size='859G',
    task='Segmentation',
)
@checksum('bimcv_covid19')
class BIMCVCovid19(Source):
    """
    BIMCV COVID-19 Dataset, CT-images only
    It includes BIMCV COVID-19 positive partition (https://arxiv.org/pdf/2006.01174.pdf)
    and negative partion
    (https://ieee-dataport.org/open-access/bimcv-covid-19-large-annotated-dataset-rx-and-ct-images-covid-19-patients-0)

    PCR tests are not used

    GitHub page: https://github.com/BIMCV-CSUSP/BIMCV-COVID-19

    Parameters
    ----------
    root : str, Path
        path to the folder containing the downloaded and parsed data.

    Notes
    -----
    Dataset has 2 partitions: bimcv-covid19-positive and bimcv-covid19-positive
    Each partition is spread over the 81 different tgz archives. The archives includes metadata about
    subject, sessions, and labels. Also there are some tgz archives for nifty images in nii.gz format

    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> ds = BIMCVCovid19(root='/path/to/downloaded/data/folder/')
    >>> print(len(ds.ids))
    # 201
    >>> print(ds.image(ds.ids[0]).shape)
    # (512, 512, 163)
    >>> print(ds.is_positive(ds.ids[0]))
    # True
    >>> print(ds.subject_info[80])
    # {'modality_dicom': "['CT']",
    #  'body_parts': "[['chest']]",
    #  'age': '[80]',
    #  'gender': 'M'}

    References
    ----------
    .. [1] Maria De La Iglesia Vayá, Jose Manuel Saborit, Joaquim Angel Montell, Antonio Pertusa, Aurelia
            Bustos, Miguel Cazorla, Joaquin Galant, Xavier Barber, Domingo Orozco-Beltrán, Francisco
            Garcia, Marisa Caparrós, Germán González, and Jose María Salinas. BIMCV COVID-19+: a
            large annotated dataset of RX and CT images from COVID-19 patients. arXiv:2006.01174, 2020.
    .. [2] Maria de la Iglesia Vayá, Jose Manuel Saborit-Torres, Joaquim Angel Montell Serrano,
            Elena Oliver-Garcia, Antonio Pertusa, Aurelia Bustos, Miguel Cazorla, Joaquin Galant,
            Xavier Barber, Domingo Orozco-Beltrán, Francisco García-García, Marisa Caparrós, Germán González,
            Jose María Salinas, 2021. BIMCV COVID-19-: a large annotated dataset of RX and CT images from COVID-19
            patients.
            Available at: https://dx.doi.org/10.21227/m4j2-ap59.
    """

    _root: str

    def _base(_root: Silent):
        return Path(_root)

    @lru_cache(None)
    def _positive_root(_base):
        return find_subroot(_base, 'covid19_posi')

    @lru_cache(None)
    def _negative_root(_base):
        return find_subroot(_base, 'covid19_neg')

    def _current_root(_meta, _positive_root, _negative_root):
        return _positive_root if _meta['is_positive'] else _negative_root

    @meta
    def ids(_series2metainfo):
        return tuple(sorted(_series2metainfo))

    def session_id(_meta):
        return _meta['session_id']

    def subject_id(_meta):
        return _meta['subject_id']

    def is_positive(_meta):
        return _meta['is_positive']

    def image(_meta, _current_root):
        with unpack(_current_root, _meta['archive_path'], _meta['image_path']) as (file, unpacked):
            if unpacked:
                array = np.asarray(nb.load(file).dataobj)
            else:
                with gzip.GzipFile(fileobj=file) as nii:
                    nii = nb.FileHolder(fileobj=nii)
                    image = nb.Nifti1Image.from_file_map({'header': nii, 'image': nii})
                    array = np.asarray(image.dataobj)

            integer = np.int16(array)
            if (array == integer).all():
                return integer

            # there are about 30 images that will visit this branch, so it's ok
            integer = np.int32(array)
            assert (array == integer).all(), np.abs(array - integer.astype(float)).max()
            return integer

    def affine(_meta, _current_root):
        with unpack(_current_root, _meta['archive_path'], _meta['image_path']) as (file, unpacked):
            if unpacked:
                return nb.load(file).affine

            with gzip.GzipFile(fileobj=file) as nii:
                nii = nb.FileHolder(fileobj=nii)
                image = nb.Nifti1Image.from_file_map({'header': nii, 'image': nii})
                return image.affine

    def tags(_meta, _current_root) -> dict:
        """
        dicom tags
        """
        if _meta['meta_path'] is None:
            return {}

        with unpack(_current_root, _meta['archive_path'], _meta['meta_path']) as (file, _):
            try:
                return parse_dicom_tags(load(file))
            except ValueError:
                return {}

    @lru_cache(None)
    def _labels_dataframe(_positive_root, _negative_root):
        with unpack(
            _positive_root, 'covid19_posi_head.tar.gz', 'covid19_posi/derivatives/labels/labels_covid_posi.tsv'
        ) as (file, _):
            pos_dataframe = pd.read_csv(file, sep='\t', index_col='ReportID').iloc[:, 1:]

        with unpack(
            _negative_root, 'covid19_neg_derivative.tar.gz', 'covid19_neg/derivatives/labels/Labels_covid_NEG_JAN21.tsv'
        ) as (file, _):
            neg_dataframe = pd.read_csv(file, sep='\t', index_col='ReportID').iloc[:, 1:]

        return pd.concat([pos_dataframe, neg_dataframe], ignore_index=False)

    def label_info(_meta, _labels_dataframe) -> dict:
        """
        labelCUIS, Report, LocalizationsCUIS etc.
        """
        session_id = _meta['session_id']

        if session_id in _labels_dataframe.index:
            return dict(_labels_dataframe.loc[session_id])
        else:
            return {}

    @lru_cache(None)
    def _subject_df(_positive_root, _negative_root):
        with unpack(_positive_root, 'covid19_posi_subjects.tar.gz', 'covid19_posi/participants.tsv') as (file, _):
            pos_data = pd.read_csv(file, sep='\t', index_col='participant')
            pos_data = pos_data[pos_data.index != 'derivatives']

        with unpack(_negative_root, 'covid19_neg_metadata.tar.gz', 'covid19_neg/participants.tsv') as (file, _):
            neg_data = pd.read_csv(file, sep='\t', index_col='participant')
            neg_data = neg_data[neg_data.index != 'derivatives']

        return pd.concat([pos_data, neg_data])

    def subject_info(_meta, _subject_df) -> dict:
        """
        modality_dicom (=[CT]), body_parts(=[chest]), age, gender
        """
        subject_id = _meta['subject_id']
        if subject_id in _subject_df.index:
            return dict(_subject_df.loc[subject_id])
        else:
            return {}

    def session_info(_meta, _current_root) -> dict:
        """
        study_date,	medical_evaluation
        """
        session_id = _meta['session_id']
        subject_id = _meta['subject_id']

        if _meta['is_positive']:
            step_sessions_tarfile_name = 'covid19_posi_sessions_tsv.tar.gz'
        else:
            step_sessions_tarfile_name = 'covid19_neg_sessions_tsv.tar.gz'

        txt_splits = load(_current_root / (step_sessions_tarfile_name + '.tar-tvf.txt')).split()
        (ses_file,) = filter(lambda x: subject_id in x, txt_splits)

        with unpack(_current_root, step_sessions_tarfile_name, ses_file) as (file, _):
            sesions_dataframe = pd.read_csv(file, sep='\t', index_col='session_id')

        if session_id in sesions_dataframe.index:
            return dict(sesions_dataframe.loc[session_id])
        return {}

    @lru_cache(None)
    def _series2metainfo(_positive_root, _negative_root):
        """
        Main function that gathers the metadata for the whole dataset.

        Returns
        -------
        A dict with {series_id : {'session_id': ...
                                'subject_id': ...,
                                'archive_path': ...,
                                'is_positive': ...,
                                'image_path': ...,
                                'meta_path': ...}}
        """
        series2metainfo = {}
        for part in _positive_root, _negative_root:
            for structure in part.glob('*part*.tar.gz.tar-tvf.txt'):
                part_filename = structure.name[: -len('.tar-tvf.txt')]
                if 'pos' in part_filename:
                    is_positive = True
                else:
                    assert 'neg' in part_filename
                    is_positive = False

                text_descr_splits = load(structure).split()
                for el_desrc in text_descr_splits:
                    # only chest ct images ('.json' and '.nii.gz' files)
                    # 'don't remove png but it should not be in cases'
                    if 'chest_ct' not in el_desrc or not el_desrc.endswith(('.nii.gz', '.json')):
                        continue

                    image_path = None
                    meta_path = None
                    if el_desrc.endswith('.nii.gz'):
                        ext = '.nii.gz'
                        image_path = el_desrc
                    # elif el_desrc.endswith(".json"):
                    else:
                        assert el_desrc.endswith('.json')
                        ext = '.json'
                        meta_path = el_desrc

                    # obtaining series id
                    series_id = str(Path(el_desrc).name)[: -len(ext)]

                    # if no such series yet
                    if series_id not in series2metainfo:
                        # obtain subject_id, session_id
                        (subject_id,) = filter(lambda x: 'sub' in x, series_id.split('_'))
                        (session_id,) = filter(lambda x: 'ses' in x, series_id.split('_'))

                        series2metainfo[series_id] = {
                            'session_id': session_id,
                            'subject_id': subject_id,
                            'archive_path': part_filename,
                            'is_positive': is_positive,
                            'image_path': None,
                            'meta_path': None,
                        }

                    if image_path is not None:
                        series2metainfo[series_id]['image_path'] = image_path

                    if meta_path is not None:
                        series2metainfo[series_id]['meta_path'] = meta_path

        return series2metainfo

    def _meta(key, _series2metainfo):
        return _series2metainfo[key]


class SpacingFromAffine(Transform):
    __inherit__ = True

    def spacing(affine):
        return nb.affines.voxel_sizes(affine)


@contextlib.contextmanager
def unpack(root: Path, archive: str, relative: str):
    unpacked = root / relative
    if unpacked.exists():
        yield unpacked, True
    else:
        with tarfile.open(root / archive) as part_file:
            yield part_file.extractfile(relative), False


def find_subroot(path: Path, name: str):
    """Finds a subfolder but in a bfs manner instead of dfs"""
    folders = [path]

    while folders:
        current = folders.pop(0)

        for child in current.iterdir():
            if child.name.startswith(name) and child.name.endswith('.tar-tvf.txt'):
                return current

            if child.is_dir():
                folders.append(child)

    raise FileNotFoundError(
        'No "*.tar-tvf.txt" files have been found. They are needed to gather the datasets structure.'
    )


def parse_dicom_tags(tags: tp.Dict[str, tp.Any]) -> tp.Optional[tp.Union[dict, list]]:
    if not isinstance(tags, dict):
        return tags
    if len(tags) == 1:
        if 'vr' in tags:
            return None
        return parse_dicom_tags(list(tags.values())[0])

    if set(tags.keys()) == {'Value', 'vr'}:
        value = tags['Value']
        # vr = tags["vr"]
        assert isinstance(value, list)
        result = [parse_dicom_tags(v) for v in value]
        if not result:
            return None
        if len(result) == 1:
            return result[0]
        return result

    result_: tp.Dict[str, tp.Optional[tp.Union[dict, list]]] = {}
    for tag, value in tags.items():
        try:
            keyword = pydicom.datadict.keyword_for_tag(tag)
        except ValueError:
            keyword = str(tag)
        assert isinstance(result_, dict)
        result_[keyword] = parse_dicom_tags(value)
    return result_



import runpy
from pathlib import Path

from setuptools import find_packages, setup


classifiers = [
    'Development Status :: 4 - Beta',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
    'Programming Language :: Python :: 3 :: Only',
]

name = 'amid'
root = Path(__file__).resolve().parent
with open(root / 'requirements.txt', encoding='utf-8') as file:
    requirements = file.read().splitlines()
with open(root / 'README.md', encoding='utf-8') as file:
    long_description = file.read()
version = runpy.run_path(root / name / '__version__.py')['__version__']

setup(
    name=name,
    packages=find_packages(include=(name,)),
    include_package_data=True,
    version=version,
    description='A curated list of medical imaging datasets with unified interfaces',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/neuro-ml/amid',
    download_url='https://github.com/neuro-ml/amid/archive/v%s.tar.gz' % version,
    keywords=['medical imaging', 'dataset'],
    entry_points={
        'bev_plugin': ['amid = amid.internals.hooks'],
        'console_scripts': ['amid = amid.internals.cli:main'],
    },
    classifiers=classifiers,
    install_requires=requirements,
    python_requires='>=3.6',
)



SQ = "'"

COLUMN2LABEL = {
    f'Patient{SQ}s Birth Date': 'birth_date',
    f'Patient{SQ}s Sex': 'sex',
    f'Patient{SQ}s Age': 'age',
    f'Manufacturer{SQ}s Model Name': 'manufacturer_model',
    'Manufacturer': 'manufacturer',
    'Acquisition Date': 'acquisition_date',
    'Site': 'site',
}



import codecs
import json
import warnings
from functools import lru_cache
from pathlib import Path

import numpy as np
import pydicom
from connectome import Source, meta
from connectome.interface.nodes import Output, Silent
from dicom_csv import (
    get_common_tag,
    get_orientation_matrix,
    get_pixel_spacing,
    get_slice_locations,
    get_tag,
    order_series,
    stack_images,
)
from dicom_csv.exceptions import TagMissingError
from tqdm.auto import tqdm

from ..internals import checksum, register
from ..utils import get_series_date
from .nodules import get_nodules


@register(
    body_region='Thorax',
    modality='CT',
    task='Lung Cancer Detection',
    link='https://mosmed.ai/en/datasets/mosmeddata-kt-s-priznakami-raka-legkogo-tip-viii/',
    prep_data_size='103G',
    raw_data_size='187G',
)
@checksum('cancer_500')
class MoscowCancer500(Source):
    """
    The Moscow Radiology Cancer-500 dataset.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded files.
        If not provided, the cache is assumed to be already populated.
    version : str, optional
        the data version. Only has effect if the library was installed from a cloned git repository.

    Notes
    -----
    Download links:
    https://mosmed.ai/en/datasets/mosmeddata-kt-s-priznakami-raka-legkogo-tip-viii/
    After pressing the `download` button you will have to provide an email address to which further instructions
    will be sent.

    Examples
    --------
    >>> # Place the downloaded files in any folder and pass the path to the constructor:
    >>> ds = MoscowCancer500(root='/path/to/files/root')
    >>> print(len(ds.ids))
    # 979
    >>> print(ds.image(ds.ids[0]).shape)
    # (512, 512, 67)
    """

    _root: str = None

    @lru_cache(None)
    def _mapping(_root: Silent):
        if _root is None:
            raise ValueError('Please pass the location of the downloaded files')

        _root = Path(_root)
        path = _root / 'series-to-files.json'
        if not path.exists():
            mapping = {}
            for file in tqdm(
                _root.rglob('*'), total=sum(1 for _ in _root.rglob('*')), desc='Analyzing folder structure'
            ):
                if file.is_dir():
                    continue

                series = pydicom.dcmread(file, specific_tags=[(0x0020, 0x000E)]).SeriesInstanceUID
                mapping[series].append(str(file.relative_to(_root)))

            with open(path, 'w') as file:
                json.dump(mapping, file)
            return mapping

        with open(path) as file:
            return json.load(file)

    @meta
    def ids(_mapping):
        # this id has an undefined image orientation
        ignore = {'1.2.643.5.1.13.13.12.2.77.8252.604378326291403.583548115656123.'}
        return tuple(sorted(set(_mapping) - ignore))

    def _series(i, _mapping, _root: Silent):
        series = [pydicom.dcmread(Path(_root, 'dicom', f)) for f in _mapping[i]]
        series = order_series(series, decreasing=False)
        return series

    def image(_series):
        x = stack_images(_series, -1).astype(np.int16)
        # DICOM specifies that the first 2 axes are (y, x). let's fix that
        return np.moveaxis(x, 0, 1)

    def study_uid(_series):
        return get_common_tag(_series, 'StudyInstanceUID')

    def series_uid(_series):
        return get_common_tag(_series, 'SeriesInstanceUID')

    def sop_uids(_series):
        return [str(get_tag(i, 'SOPInstanceUID')) for i in _series]

    def pixel_spacing(_series):
        return get_pixel_spacing(_series).tolist()

    def slice_locations(_series):
        return get_slice_locations(_series)

    def orientation_matrix(_series):
        return get_orientation_matrix(_series)

    def instance_numbers(_series):
        try:
            instance_numbers = [int(get_tag(i, 'InstanceNumber')) for i in _series]
            if not _is_monotonic(instance_numbers):
                warnings.warn('Ordered series has non-monotonic instance numbers.')

            return instance_numbers
        except TagMissingError:
            pass

    def conv_kernel(_series):
        return get_common_tag(_series, 'ConvolutionKernel', default=None)

    def kvp(_series):
        return get_common_tag(_series, 'KVP', default=None)

    def patient_id(_series):
        return get_common_tag(_series, 'PatientID', default=None)

    def study_date(_series):
        return get_series_date(_series)

    def accession_number(_series):
        return get_common_tag(_series, 'AccessionNumber', default=None)

    def nodules(i, _mapping, _root: Silent, _series, slice_locations: Output):
        folders = {Path(f).parent.name for f in _mapping[i]}
        if len(folders) != 1:
            # can't determine protocol filename
            return

        (filename,) = folders
        protocol = json.load(codecs.open(str(Path(_root) / 'protocols' / f'{filename}.json'), 'r', 'utf-8-sig'))

        series_number = get_common_tag(_series, 'SeriesNumber')
        try:
            return get_nodules(protocol, series_number, slice_locations)
        except ValueError:
            pass


def _is_monotonic(sequence):
    sequence = list(sequence)
    return sequence == sorted(sequence) or sequence == sorted(sequence)[::-1]



from functools import lru_cache
from pathlib import Path

import nibabel as nb
import numpy as np
import pandas as pd
from connectome import Output, Source, meta
from connectome.interface.nodes import Silent

from ..internals import checksum, licenses, register
from .data_classes import AcquisitionInfo, ClinicalInfo


@register(
    body_region='Head',
    license=licenses.CC_BY_40,
    link='https://wiki.cancerimagingarchive.net/pages/viewpage.action?pageId=70225642',
    modality=('FLAIR', 'MRI T1', 'MRI T1GD', 'MRI T2', 'DSC MRI', 'DTI MRI'),
    prep_data_size='70G',
    raw_data_size='69G',
    task='Segmentation',
)
@checksum('upenn_gbm')
class UPENN_GBM(Source):
    """
    Multi-parametric magnetic resonance imaging (mpMRI) scans for de novo Glioblastoma
      (GBM) patients from the University of Pennsylvania Health System (UPENN-GBM).
    Dataset contains 630 patients.

    All samples are registered to a common atlas (SRI)
        using a uniform preprocessing and the segmentation are aligned with them.


    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.

    Notes
    -----
    Follow the download instructions at https://wiki.cancerimagingarchive.net/pages/viewpage.action?pageId=70225642
    Download to the root folder nifti images and metadata. Organise folder as folows:


    <...>/<UPENN-root>/NIfTI-files/images_segm/UPENN-GBM-00054_11_segm.nii.gz
    <...>/<UPENN-root>/NIfTI-files/...

    <...>/<UPENN-root>/UPENN-GBM_clinical_info_v1.0.csv
    <...>/<UPENN-root>/UPENN-GBM_acquisition.csv


    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> ds = UPENN_GBM(root='/path/to/downloaded/data/folder/')
    >>> print(len(ds.ids))
    # 671
    >>> print(ds.image(ds.ids[215]).shape)
    # (4, 240, 240, 155)
    >>> print(d.acqusition_info(d.ids[215]).manufacturer)
    # SIEMENS

    References
    ----------
    .. [1] Bakas, S., Sako, C., Akbari, H., Bilello, M., Sotiras, A., Shukla, G., Rudie,
      J. D., Flores Santamaria, N., Fathi Kazerooni, A., Pati, S., Rathore, S.,
    Mamourian, E., Ha, S. M., Parker, W., Doshi, J., Baid, U., Bergman, M., Binder, Z. A., Verma, R., … Davatzikos,
    C. (2021). Multi-parametric magnetic resonance imaging (mpMRI) scans for de novo
    Glioblastoma (GBM) patients from the University of Pennsylvania Health System (UPENN-GBM)
    (Version 2) [Data set]. The Cancer Imaging Archive.
    https://doi.org/10.7937/TCIA.709X-DN49

    """

    _root: str = None

    def _base(_root: Silent):
        if _root is None:
            raise ValueError('Please provide the `root` argument')
        return Path(_root)

    @meta
    def ids(_base):
        ids = [x.name for x in (_base / 'NIfTI-files/images_structural').iterdir()]
        return tuple(sorted(ids))

    def modalities():
        return ['T1', 'T1GD', 'T2', 'FLAIR']

    def dsc_modalities():
        return ['', 'ap-rCBV', 'PH', 'PSR']

    def dti_modalities():
        return ['AD', 'FA', 'RD', 'TR']

    def _mask_path(i, _base):
        p1 = _base / 'NIfTI-files/images_segm'
        p2 = _base / 'NIfTI-files/automated_segm'
        p1 = list(p1.glob(i + '*'))
        p2 = list(p2.glob(i + '*'))
        return p1[0] if p1 else p2[0] if p2 else None

    def mask(_mask_path: Path):
        if not _mask_path:
            return None
        return np.asarray(nb.load(_mask_path).get_fdata())

    def is_mask_automated(_mask_path: Path):
        if _mask_path is None:
            return None
        return _mask_path.parent.name == 'automated_segm'

    def image(i, modalities: Output, _base):
        path = _base / f'NIfTI-files/images_structural/{i}'
        image_pathes = [path / f'{i}_{mod}.nii.gz' for mod in modalities]
        images = [np.asarray(nb.load(p).dataobj) for p in image_pathes]
        return np.stack(images)

    def image_unstripped(i, modalities: Output, _base):
        path = _base / f'NIfTI-files/images_structural_unstripped/{i}'
        image_pathes = [path / f'{i}_{mod}_unstripped.nii.gz' for mod in modalities]
        images = [np.asarray(nb.load(p).dataobj) for p in image_pathes]
        return np.stack(images)

    def image_DTI(i, dti_modalities: Output, _base):
        path = _base / f'NIfTI-files/images_DTI/{i}'
        if not path.exists():
            return None
        image_pathes = [path / f'{i}_DTI_{mod}.nii.gz' for mod in dti_modalities]
        images = [np.asarray(nb.load(p).dataobj) for p in image_pathes]
        return np.stack(images)

    def image_DSC(i, dsc_modalities: Output, _base):
        path = _base / f'NIfTI-files/images_DSC/{i}'
        if not path.exists():
            return None
        image_pathes = [path / (f'{i}_DSC_{mod}.nii.gz' if mod else f'{i}_DSC.nii.gz') for mod in dsc_modalities]
        images = [np.asarray(nb.load(p).dataobj) for p in image_pathes]
        return images

    @lru_cache(1)
    def _clinical_info(_base):
        return pd.read_csv(_base / 'UPENN-GBM_clinical_info_v1.0.csv')

    @lru_cache(1)
    def _acqusition_info(_base):
        return pd.read_csv(_base / 'UPENN-GBM_acquisition.csv')

    def clinical_info(i, _clinical_info) -> ClinicalInfo:
        row = _clinical_info[_clinical_info.ID == i]
        return ClinicalInfo(*row.iloc[0, 1:])

    def acqusition_info(i, _acqusition_info) -> AcquisitionInfo:
        row = _acqusition_info[_acqusition_info.ID == i]
        return AcquisitionInfo(*row.iloc[0, 1:])

    def subject_id(i):
        return i.split('_')[0]

    def affine(i):
        return np.array([[-1.0, 0.0, 0.0, -0.0], [0.0, -1.0, 0.0, 239.0], [0.0, 0.0, 1.0, 0.0], [0.0, 0.0, 0.0, 1.0]])

    def spacing(i):
        return (1, 1, 1)



import contextlib
import gzip
import zipfile
from pathlib import Path
from zipfile import ZipFile

import nibabel as nb
import numpy as np
from connectome import Output, Source, meta
from connectome.interface.nodes import Silent

from .internals import checksum, licenses, register
from .utils import deprecate


@register(
    body_region='Chest',
    license=licenses.CC0_10,
    link='http://medicalsegmentation.com/covid19/',
    modality='CT',
    prep_data_size='300M',
    raw_data_size='310M',
    task='COVID-19 segmentation',
)
@checksum('medseg9')
class Medseg9(Source):
    """

    Medseg9 is a public COVID-19 CT segmentation dataset with 9 annotated images.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.
    version : str, optional
        the data version. Only has effect if the library was installed from a cloned git repository.

    Notes
    -----
    Data can be downloaded here: http://medicalsegmentation.com/covid19/.

    Then, the folder with raw downloaded data should contain three zip archives with data and masks
    (`rp_im.zip`, `rp_lung_msk.zip`, `rp_msk.zip`).

    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> ds = Medseg9(root='/path/to/downloaded/data/folder/')
    >>> print(len(ds.ids))
    # 9
    >>> print(ds.image(ds.ids[0]).shape)
    # (630, 630, 45)
    >>> print(ds.covid(ds.ids[0]).shape)
    # (630, 630, 45)

    """

    _root: str = None

    @meta
    def ids(_root: Silent):
        result = set()

        with ZipFile(Path(_root) / 'rp_msk.zip') as zf:
            for zipinfo in zf.infolist():
                if zipinfo.is_dir():
                    continue
                file_stem = Path(zipinfo.filename).stem
                result.add('medseg9_' + file_stem.split('.nii')[0])

        return tuple(sorted(result))

    def _file(i, _root: Silent):
        num_id = i.split('_')[-1]
        return zipfile.Path(Path(_root) / 'rp_im.zip', f'rp_im/{num_id}.nii.gz')

    def image(_file):
        with open_nii_gz_file(_file) as nii_image:
            # most CT/MRI scans are integer-valued, this will help us improve compression rates
            return np.int16(nii_image.get_fdata())

    def affine(_file):
        """The 4x4 matrix that gives the image's spatial orientation."""
        with open_nii_gz_file(_file) as nii_image:
            return nii_image.affine

    @deprecate(message='Use `spacing` method instead.')
    def voxel_spacing(spacing: Output):
        return spacing

    def spacing(_file):
        """Returns voxel spacing along axes (x, y, z)."""
        with open_nii_gz_file(_file) as nii_image:
            return tuple(nii_image.header['pixdim'][1:4])

    def lungs(_file):
        mask_file = zipfile.Path(Path(_file.root.filename).parent / 'rp_lung_msk.zip', f'rp_lung_msk/{_file.name}')
        with open_nii_gz_file(mask_file) as nii_image:
            return np.bool_(nii_image.get_fdata())

    def covid(_file):
        """
        int16 mask.
        0 - normal, 1 - ground-glass opacities (матовое стекло), 2 - consolidation (консолидация).
        """
        mask_file = zipfile.Path(Path(_file.root.filename).parent / 'rp_msk.zip', f'rp_msk/{_file.name}')
        with open_nii_gz_file(mask_file) as nii_image:
            # most CT/MRI scans are integer-valued, this will help us improve compression rates
            return np.uint8(nii_image.get_fdata())


# TODO: sync with amid.utils
@contextlib.contextmanager
def open_nii_gz_file(file):
    with file.open('rb') as opened:
        with gzip.GzipFile(fileobj=opened) as nii:
            nii = nb.FileHolder(fileobj=nii)
            yield nb.Nifti1Image.from_file_map({'header': nii, 'image': nii})



from gzip import GzipFile
from pathlib import Path
from typing import Dict, Sequence, Union

import numpy as np
from bev import Repository
from connectome import CacheColumns as Columns, CacheToDisk as Disk
from connectome.utils import StringsLike
from tarn import HashKeyStorage, ReadError
from tarn.serializers import (
    ChainSerializer,
    DictSerializer,
    JsonSerializer as BaseJsonSerializer,
    PickleSerializer,
    Serializer,
    SerializerError,
)


class CacheToDisk(Disk):
    def __init__(
        self,
        names: StringsLike,
        serializer: Union[Serializer, Sequence[Serializer]] = None,
        **kwargs,
    ):
        repo = Repository.from_here('../data')
        cache = repo.cache
        super().__init__(
            [x.root for x in cache.local[0].locations],
            cache.storage,
            serializer=default_serializer(serializer),
            names=names,
            **kwargs,
        )


class CacheColumns(Columns):
    def __init__(
        self,
        names: StringsLike,
        serializer: Union[Serializer, Sequence[Serializer]] = None,
        **kwargs,
    ):
        repo = Repository.from_here('../data')
        cache = repo.cache
        super().__init__(
            [x.root for x in cache.local[0].locations],
            cache.storage,
            serializer=default_serializer(serializer),
            names=names,
            **kwargs,
        )


def default_serializer(serializers):
    if serializers is None:
        arrays = NumpySerializer({np.bool_: 1, np.integer: 1})
        serializers = ChainSerializer(
            JsonSerializer(),
            DictSerializer(serializer=arrays),
            arrays,
            PickleSerializer(),
            # CleanInvalid()
        )
    return serializers


class NumpySerializer(Serializer):
    def __init__(self, compression: Union[int, Dict[type, int]] = None):
        self.compression = compression

    def _choose_compression(self, value):
        if isinstance(self.compression, int) or self.compression is None:
            return self.compression

        if isinstance(self.compression, dict):
            for dtype in self.compression:
                if np.issubdtype(value.dtype, dtype):
                    return self.compression[dtype]

    def save(self, value, folder: Path):
        if not isinstance(value, (np.ndarray, np.generic)):
            raise SerializerError

        compression = self._choose_compression(value)
        if compression is not None:
            assert isinstance(compression, int)
            with GzipFile(folder / 'value.npy.gz', 'wb', compresslevel=compression, mtime=0) as file:
                np.save(file, value, allow_pickle=False)

        else:
            np.save(folder / 'value.npy', value, allow_pickle=False)

    def load(self, folder: Path, storage: HashKeyStorage):
        paths = list(folder.iterdir())
        if len(paths) != 1:
            raise SerializerError

        (path,) = paths
        if path.name == 'value.npy':
            loader = np.load
        elif path.name == 'value.npy.gz':

            def loader(x):
                with GzipFile(x, 'rb') as file:
                    return np.load(file)

        else:
            raise SerializerError

        try:
            return self._load_file(storage, loader, path)
        except (ValueError, EOFError) as e:
            raise ReadError from e


class JsonSerializer(BaseJsonSerializer):
    def save(self, value, folder: Path):
        # if namedtuple
        if isinstance(value, tuple) and hasattr(value, '_asdict') and hasattr(value, '_fields'):
            raise SerializerError

        return super().save(value, folder)


class CleanInvalid(Serializer):
    def save(self, value, folder: Path):
        raise SerializerError

    def load(self, folder: Path, storage: HashKeyStorage):
        raise ReadError
